# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Importing Framework
"""

from __future__ import unicode_literals, absolute_import

from .importers import Importer, FromCSV, FromQuery, FromDjango, BatchImporter, BulkImporter
from .sqlalchemy import FromSQLAlchemy, ToSQLAlchemy
from .postgresql import BulkToPostgreSQL
from .handlers import ImportHandler, BulkImportHandler, FromSQLAlchemyHandler, ToSQLAlchemyHandler
from .rattail import FromRattailHandler, ToRattailHandler
from . import model



# -*- coding: utf-8; -*-
"""grow ProductBatch.item_entry

Revision ID: dfc1ed686f3f
Revises: 9ec8deec5256
Create Date: 2020-03-01 14:14:22.198396

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'dfc1ed686f3f'
down_revision = '9ec8deec5256'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch*
    op.alter_column('batch_handheld_row', 'item_entry', type_=sa.String(length=32))
    op.alter_column('batch_inventory_row', 'item_entry', type_=sa.String(length=32))
    op.alter_column('batch_newproduct_row', 'item_entry', type_=sa.String(length=32))
    op.alter_column('batch_pricing_row', 'item_entry', type_=sa.String(length=32))
    op.alter_column('batch_product_row', 'item_entry', type_=sa.String(length=32))
    op.alter_column('label_batch_row', 'item_entry', type_=sa.String(length=32))
    op.alter_column('purchase_batch_row', 'item_entry', type_=sa.String(length=32))
    op.alter_column('vendor_catalog_row', 'item_entry', type_=sa.String(length=32))
    op.alter_column('vendor_invoice_row', 'item_entry', type_=sa.String(length=32))


def downgrade():

    # batch*
    op.alter_column('vendor_invoice_row', 'item_entry', type_=sa.String(length=20))
    op.alter_column('vendor_catalog_row', 'item_entry', type_=sa.String(length=20))
    op.alter_column('purchase_batch_row', 'item_entry', type_=sa.String(length=20))
    op.alter_column('label_batch_row', 'item_entry', type_=sa.String(length=20))
    op.alter_column('batch_product_row', 'item_entry', type_=sa.String(length=20))
    op.alter_column('batch_pricing_row', 'item_entry', type_=sa.String(length=20))
    op.alter_column('batch_newproduct_row', 'item_entry', type_=sa.String(length=20))
    op.alter_column('batch_inventory_row', 'item_entry', type_=sa.String(length=20))
    op.alter_column('batch_handheld_row', 'item_entry', type_=sa.String(length=20))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendors Handler
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail.app import GenericHandler
from rattail.util import load_entry_points


class VendorHandler(GenericHandler):
    """
    Base class and default implementation for vendor handlers.
    """

    def choice_uses_dropdown(self):
        """
        Returns boolean indicating whether a vendor choice should be
        presented to the user via a dropdown (select) element, vs.  an
        autocomplete field.  The latter is the default because
        potentially the vendor list can be quite large, so we avoid
        loading them all in the dropdown unless so configured.

        :returns: Boolean; if true then a dropdown should be used;
           otherwise (false) autocomplete is used.
        """
        return self.config.getbool('rattail', 'vendors.choice_uses_dropdown',
                                   default=False)

    def get_vendor(self, session, key, **kwargs):
        """
        Locate and return the vendor corresponding to the given key.

        The key can be a UUID value, but most often it will instead be
        a "generic" key specific to this purpose.  Any generic key can
        be defined within the settings, pointing to a valid vendor.
        
        For instance, we can define a key of ``'poser.acme'`` to
        denote the hypothetical "Acme Distribution" vendor, and we add
        a namespace unique to our app just to be safe.

        We then create a setting in the DB pointing to our *actual*
        vendor by way of its UUID:

        .. code-block:: sql

           INSERT INTO SETTING (name, value) 
           VALUES ('rattail.vendor.poser.acme',
                   '7e6d69a2700911ec93533ca9f40bc550');

        From then on we could easily fetch the vendor by this key.
        This is mainly useful to allow catalog and invoice parsers to
        "loosely" associate with a particular vendor by way of this
        key, which could be shared across organizations etc.
        """
        from rattail.db.api.vendors import get_vendor
        return get_vendor(session, key)

    def render_vendor(self, vendor, **kwargs):
        return six.text_type(vendor)

    def get_all_catalog_parsers(self):
        """
        Should return *all* catalog parsers known to exist.

        Note that this returns classes and not instances.

        :returns: List of
           :class:`~rattail.vendors.catalogs.CatalogParser` classes.
        """
        Parsers = list(
            load_entry_points('rattail.vendors.catalogs.parsers').values())
        Parsers.sort(key=lambda Parser: Parser.display)
        return Parsers
        
    def get_supported_catalog_parsers(self):
        """
        Should return only those catalog parsers which are "supported"
        by the current app.  Usually "supported" just means what we
        want to expose to the user.

        Note that this returns classes and not instances.

        :returns: List of
           :class:`~rattail.vendors.catalogs.CatalogParser` classes.
        """
        Parsers = self.get_all_catalog_parsers()
        
        supported_keys = self.config.getlist(
            'rattail', 'vendors.supported_catalog_parsers')
        if supported_keys is None:
            supported_keys = self.config.getlist(
                'tailbone', 'batch.vendorcatalog.supported_parsers')
        if supported_keys:
            Parsers = [Parser for Parser in Parsers
                       if Parser.key in supported_keys]
            
        return Parsers

    def get_catalog_parser(self, key, require=False):
        """
        Retrieve the catalog parser for the given parser key.

        Note that this returns an instance, not the class.

        :param key: Unique key indicating which parser to get.

        :returns: A :class:`~rattail.vendors.catalogs.CatalogParser`
           instance.
        """
        from rattail.vendors.catalogs import CatalogParserNotFound

        for Parser in self.get_all_catalog_parsers():
            if Parser.key == key:
                return Parser(self.config)

        if require:
            raise CatalogParserNotFound(key)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Customer Orders Handler

Please note this is different from the Customer Order Batch Handler.
"""

from __future__ import unicode_literals, absolute_import

from rattail.app import GenericHandler


class CustomerOrderHandler(GenericHandler):
    """
    Base class and default implementation for customer order handlers.
    """

    def resolve_person(self, pending, person, user, **kwargs):
        """
        Resolve a pending person for all customer orders.
        """
        for order in list(pending.custorder_records):
            order.person = person
            order.pending_customer = None
            for item in order.items:
                item.add_event(self.enum.CUSTORDER_ITEM_EVENT_CUSTOMER_RESOLVED,
                               user)

    def resolve_product(self, pending, product, user, **kwargs):
        """
        Resolve a pending product for all customer orders.
        """
        for item in pending.custorder_item_records:
            item.product = product
            item.pending_product = None

            item.product_upc = product.upc
            item.product_item_id = product.item_id
            item.product_scancode = product.scancode
            item.product_brand = product.brand.name if product.brand else None
            item.product_description = product.description
            item.product_size = product.size

            # TODO: not sure this is needed really?
            item.product_weighed = product.weighed

            # TODO: model notes say this is not needed
            #item.product_unit_of_measure = product.unit_of_measure

            department = product.department
            item.department_number = department.number if department else None
            item.department_name = department.name if department else None

            # TODO: should be smarter about getting this
            item.case_quantity = product.case_size

            cost = product.cost
            item.product_unit_cost = cost.unit_cost if cost else None

            regprice = product.regular_price
            item.unit_regular_price = regprice.price if regprice else None

            curprice = product.current_price
            item.unit_sale_price = curprice.price if curprice else None
            item.sale_ends = curprice.ends if curprice else None

            item.unit_price = item.unit_sale_price or item.unit_regular_price

            # TODO: should recalculate total price
            #item.total_price = ...

            item.add_event(self.enum.CUSTORDER_ITEM_EVENT_PRODUCT_RESOLVED,
                           user)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
API for Organizational Models
"""

from __future__ import unicode_literals, absolute_import

import six
from sqlalchemy.orm.exc import NoResultFound

from rattail.db import model


# TODO: need to refactor this into the AppHandler, or..?
def get_department(session, key):
    """
    Locate and return a department for the given key, if possible.

    First the key is assumed to be a ``Department.number`` value.  If no
    matches are found, then it looks for a special setting in the database.  If
    one is found, ``get_department()`` is called again with its value.

    :param session: Active database session.

    :param key: Value to use when searching for the department.  

    :returns: The :class:`rattail.db.model.Department` instance if found;
      otherwise ``None``.
    """
    # Department.uuid match?
    department = session.query(model.Department).get(six.text_type(key))
    if department:
        return department

    # Department.number match?
    if isinstance(key, int) or key.isdigit():
        try:
            return session.query(model.Department).filter_by(number=key).one()
        except NoResultFound:
            pass

    # Try settings, if value then recurse.
    from rattail.db.api import get_setting
    key = get_setting(session, 'rattail.department.{0}'.format(key))
    if key is None:
        return None
    return get_department(session, key)


# TODO: need to refactor this into the AppHandler, or..?
def get_subdepartment(session, key):
    """
    Locate and return a subdepartment for the given key, if possible.

    First the key is assumed to be a ``Subdepartment.number`` value.  If no
    matches are found, then it looks for a special setting in the database.  If
    one is found, ``get_subdepartment()`` is called again with its value.

    :param session: Active database session.

    :param key: Value to use when searching for the subdepartment.  

    :returns: The :class:`rattail.db.model.Subdepartment` instance if found;
      otherwise ``None``.
    """
    # Subdepartment.uuid match?
    subdepartment = session.query(model.Subdepartment).get(key)
    if subdepartment:
        return subdepartment

    # Subdepartment.number match?
    if isinstance(key, int) or key.isdigit():
        try:
            return session.query(model.Subdepartment).filter_by(number=key).one()
        except NoResultFound:
            pass

    # Try settings, if value then recurse.
    from rattail.db.api import get_setting
    key = get_setting(session, 'rattail.subdepartment.{0}'.format(key))
    if key is None:
        return None
    return get_subdepartment(session, key)



# -*- coding: utf-8 -*-
"""increase vendor name length

Revision ID: 2e3f33b8ea46
Revises: 1515aa76b64a
Create Date: 2015-02-05 11:29:53.537647

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '2e3f33b8ea46'
down_revision = u'1515aa76b64a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():
    op.alter_column('vendors', 'name', type_=sa.String(length=50))


def downgrade():
    op.alter_column('vendors', 'name', type_=sa.String(length=40))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Membership Handler
"""

from __future__ import unicode_literals, absolute_import

from rattail.util import load_object
from rattail.app import GenericHandler


class MembershipHandler(GenericHandler):
    """
    Base class and default implementation for membership handlers.
    """

    def ensure_member(self, person, **kwargs):
        """
        Returns the Member record associated with the given person, creating
        it first if necessary.
        """
        raise NotImplementedError

    def make_member(self, person, **kwargs):
        """
        Make and return a new Member instance.
        """
        raise NotImplementedError

    def begin_membership(self, member, **kwargs):
        """
        Begin an active membership.
        """
        raise NotImplementedError

    def get_member(self, person):
        """
        Returns the member associated with the given person, if there is one.
        """
        raise NotImplementedError

    def get_customer(self, member):
        """
        Returns the customer associated with the given member, if there is one.
        """
        raise NotImplementedError

    def get_person(self, member):
        """
        Returns the person associated with the given member, if there is one.
        """
        clientele = self.app.get_clientele_handler()
        customer = self.get_customer(member)
        person = clientele.get_person(customer)
        return person

    def get_last_patronage_date(self, member, **kwargs):
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""make DepositLink.code a string

Revision ID: 09ed5bdc99f1
Revises: c2af678f010c
Create Date: 2017-02-10 00:00:49.705475

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '09ed5bdc99f1'
down_revision = u'c2af678f010c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # deposit_link
    op.alter_column('deposit_link', 'code', type_=sa.String(length=20))


def downgrade():

    # deposit_link
    op.alter_column('deposit_link', 'code', type_=sa.Integer())



# -*- coding: utf-8 -*-
"""add person notes

Revision ID: 23600f00cda7
Revises: 7f88949a7ab9
Create Date: 2017-10-17 12:13:12.135529

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '23600f00cda7'
down_revision = u'7f88949a7ab9'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # note
    op.create_table('note',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('parent_type', sa.String(length=20), nullable=False),
                    sa.Column('parent_uuid', sa.String(length=32), nullable=False),
                    sa.Column('type', sa.String(length=15), nullable=True),
                    sa.Column('subject', sa.String(length=255), nullable=True),
                    sa.Column('text', sa.Text(), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'note_fk_created_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('note_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('type', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('subject', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('text', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('created', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('created_by_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_note_version_end_transaction_id'), 'note_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_note_version_operation_type'), 'note_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_note_version_transaction_id'), 'note_version', ['transaction_id'], unique=False)


def downgrade():

    # note
    op.drop_index(op.f('ix_note_version_transaction_id'), table_name='note_version')
    op.drop_index(op.f('ix_note_version_operation_type'), table_name='note_version')
    op.drop_index(op.f('ix_note_version_end_transaction_id'), table_name='note_version')
    op.drop_table('note_version')
    op.drop_table('note')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync Consumers
"""

from __future__ import unicode_literals, absolute_import

import warnings

import six

from rattail import importing
from rattail.db import Session
from rattail.config import parse_list
from rattail.util import load_object


class DataSyncConsumer(object):
    """
    Base class for all DataSync consumers.
    """
    delay = 1 # seconds
    retry_attempts = 1
    retry_delay = 1 # seconds

    def __init__(self, config, key, dbkey=None, runas=None, watcher=None):
        self.config = config
        self.key = key
        self.dbkey = dbkey
        self.runas_username = runas
        self.watcher = watcher
        self.model = config.get_model()
        self.enum = config.get_enum()
        self.app = config.get_app()

    def setup(self):
        """
        This method is called when the consumer thread is first started.
        """

    def begin_transaction(self):
        """
        Called just before the consumer is asked to process changes, possibly
        via multiple batches.
        """

    def pre_process_changes(self, session, changes):
        """
        Implement any "pre-processing" logic here.

        Common example would be to establish DB connection to some
        external system(s).

        In some rare cases you may also need to "sort" the changes, to
        ensure they are processed in a particular sequence.

        :returns: If you needed to modify the list of changes, you can
           return the new list.  If this method returns ``None``
           (which the default logic does) then the original set of
           changes will be processed.
        """
        # declare responsible user
        if self.runas_username:
            session.set_continuum_user(self.runas_username)

        # update all importers with current Rattail session
        for importer in six.itervalues(self.importers):
            importer.session = session

    def process_changes(self, session, changes):
        """
        Process (consume) a batch of changes.
        """

    def post_process_changes(self, session, changes):
        """
        Implement any "post-processing" logic here.

        Common example would be to close DB connection for external
        system(s).
        """

    def rollback_transaction(self):
        """
        Called when any batch of changes failed to process.
        """

    def commit_transaction(self):
        """
        Called just after the consumer has successfully finished processing
        changes, possibly via multiple batches.
        """


class DataSyncImportConsumer(DataSyncConsumer):
    """
    Base class for any DataSync Consumer which leverages an Import
    Handler (and ultimately its Importers) to do the heavy lifting.
    See :doc:`rattail-manual:data/importing/index` for more about
    Importers etc.

    .. attribute:: handler

       Reference to the import/export handler to be leveraged by this
       datasync consumer.

       The handler is automatically instantiated when the consumer is
       instantiated, usually based on :attr:`handler_key`.

    .. attribute:: handler_key

       This should be the "key" of the import handler to be leveraged
       by the datasync consumer.  This will be a string designating
       the data target/source/direction, for example
       ``'to_rattail.from_corepos_api.import'``.

       The handler itself is obtained by invoking
       :meth:`rattail.app.AppHandler.get_import_handler()`.

       Note that for now, ``handler_key`` is not strictly required, as
       :attr:`handler_spec` will be used as a fallback.  But
       eventually that will likely change, with ``handler_spec`` being
       deprecated altogether and ``handler_key`` becoming required
       (unless you override :meth:`get_handler()`).

    .. attribute:: handler_spec

       This attribute should be considered deprecated; please use
       :attr:`handler_key` instead.

       If this attribute is used, it should be a "spec" string
       referencing the import handler class from which the importers
       should be obtained.

    .. attribute:: default_importers_only

       Flag indicating whether the consumer should *only* use "default"
       importers from the handler, vs. using "all" importers, when processing
       changes.  This flag is ``True`` by default, meaning only "default"
       importers will be used.

    .. attribute:: model_map

       This is a dictionary which may be used to map model names (importer
       "keys") between the host and local systems.  Keys of the dictionary
       should be the model name as it comes from the host system, in the form
       of :attr:`rattail.db.model.DataSyncChange.payload_type`.  Values of the
       dictionary should be the model name as it is found in the import
       handler, i.e. in its ``importer_keys()`` return value.

    .. attribute:: skip_local_models

       This is a list of model names, as they are found in the import handler.
    """
    handler_key = None

    # TODO: deprecate / remove this
    handler_spec = None

    default_importers_only = True
    model_map = {}
    skip_local_models = []

    def __init__(self, *args, **kwargs):
        super(DataSyncImportConsumer, self).__init__(*args, **kwargs)

        spec = kwargs.pop('handler_spec', None)
        self.handler = self.get_handler(spec)
        self.importers = self.get_importers()

    def get_handler(self, spec=None):
        """
        Return the appropriate import handler instance.

        If ``spec`` is provided, handler will conform to that.
        Otherwise the "designated" handler is got according to
        :attr:`handler_key`, by invoking
        :meth:`rattail.app.AppHandler.get_import_handler()`.  Or if
        ``handler_key`` is not set, then for now, :attr:`handler_spec`
        may be used as fallback.

        An error of some sort will be raised if no handler can be got.

        :returns: The import handler instance.
        """
        if not spec:
            if self.handler_key:
                return self.app.get_import_handler(self.handler_key,
                                                   require=True)
            spec = self.handler_spec
        factory = self.app.load_object(spec)
        return factory(self.config)

    def get_importers(self):
        """
        Returns a dictionary, keys of which are "model" names
        (e.g. ``'Product'``) and values of which are
        :class:`~rattail.importing.importers.Importer` instances.

        .. note::
           The keys must ultimately align with the
           :attr:`~rattail.db.model.datasync.DataSyncChange.payload_type`
           values, coming from the host system.  Or at least that's what will
           make life easy for the :meth:`invoke_importer()` method.
        """
        importers = self.get_importers_from_handler(
            self.handler, default_only=self.default_importers_only)

        for host_name, local_name in self.model_map.items():
            if local_name in importers:
                importers[host_name] = importers[local_name]

        if self.skip_local_models:
            for name in list(importers):
                if name in self.skip_local_models:
                    del importers[name]

        return importers

    def get_importers_from_handler(self, handler, default_only=True):
        if default_only:
            keys = handler.get_default_keys()
        else:
            keys = handler.get_importer_keys()

        importers = dict([(key, handler.get_importer(key))
                          for key in keys])
        return importers

    def process_changes(self, session, changes):
        """
        Process all changes, leveraging importer(s) as much as possible.
        """
        result = self.pre_process_changes(session, changes)
        if result is not None:
            changes = result

        for change in changes:
            self.invoke_importer(session, change)

        self.post_process_changes(session, changes)

    def invoke_importer(self, session, change):
        """
        For the given change, invoke the default importer behavior, if one is
        available.
        """
        importer = self.importers.get(change.payload_type)
        if importer:
            if not change.deletion:
                return self.process_change(session, importer, change)
            elif importer.allow_delete:
                return self.process_deletion(session, importer, change)

    def process_change(self, session, importer, change=None, host_object=None, host_data=None):
        """
        Invoke the importer to process the given change / host record.
        """
        if host_data is None:
            if host_object is None:
                host_object = self.get_host_object(session, change)
                if host_object is None:
                    return
            host_data = importer.normalize_host_object(host_object)
            if host_data is None:
                return
        key = importer.get_key(host_data)
        local_object = self.get_local_object(importer, key, host_data)
        if local_object:
            if importer.allow_update:
                local_data = importer.normalize_local_object(local_object)
                if importer.data_diffs(local_data, host_data) and importer.allow_update:
                    local_object = importer.update_object(local_object, host_data, local_data)
            return local_object
        elif importer.allow_create:
            return importer.create_object(key, host_data)

    def process_deletion(self, session, importer, change):
        """
        Attempt to invoke the importer, to delete a local record according to
        the change involved.
        """
        if importer.allow_delete:
            key = self.get_deletion_key(session, change)
            if key is not None:
                local_object = importer.get_local_object(key)
                if local_object:
                    local_data = importer.normalize_local_object(local_object)
                    if local_data and importer.can_delete_object(
                            local_object, local_data):
                        return importer.delete_object(local_object)
        return False

    def get_deletion_key(self, session, change):
        return (change.payload_key,)

    def get_host_object(self, session, change):
        """
        You must override this, to return a host object from the given
        ``DataSyncChange`` instance.  Note that the host object need *not* be
        normalized, as that will be done by the importer.  (This is effectively
        the only part of the processing which is not handled by the importer.)
        """
        raise NotImplementedError

    def get_local_object(self, importer, key, host_data):
        """
        Fetch the "local" (destination) object for the given key.
        """
        return importer.get_local_object(key)


class NewDataSyncImportConsumer(DataSyncImportConsumer):

    def __init__(self, *args, **kwargs):
        super(NewDataSyncImportConsumer, self).__init__(*args, **kwargs)
        warnings.warn("class is deprecated, please use "
                      "DataSyncImportConsumder instead",
                      DeprecationWarning)


class FromRattailConsumer(DataSyncImportConsumer):
    """
    Base class for consumers which get their data from Rattail.
    """

    def get_host_object(self, session, change):
        model = self.model
        return session.query(getattr(model, change.payload_type))\
                      .get(change.payload_key)


class NullTestConsumer(DataSyncConsumer):
    """
    Consumer which simply does ignores changes, letting DataSync think they
    were processed okay.
    """

    def process_changes(self, session, changes):
        pass


class ErrorTestConsumer(DataSyncConsumer):
    """
    Consumer which always raises an error when processing changes.  Useful for
    testing error handling etc.
    """

    def process_changes(self, session, changes):
        raise RuntimeError("Fake exception, to test error handling")



# -*- coding: utf-8 -*-
"""add customer.active_in_pos

Revision ID: b9658d212053
Revises: 195ceb6abeb3
Create Date: 2017-05-12 14:38:44.460118

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'b9658d212053'
down_revision = u'195ceb6abeb3'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # customer
    op.add_column('customer', sa.Column('active_in_pos', sa.Boolean(), nullable=True))


def downgrade():

    # customer
    op.drop_column('customer', 'active_in_pos')



# -*- coding: utf-8; -*-
"""add patronage

Revision ID: c3b721c34f4b
Revises: 08dcf4eebe66
Create Date: 2022-02-18 18:06:06.533616

"""

from __future__ import unicode_literals, absolute_import

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c3b721c34f4b'
down_revision = '08dcf4eebe66'
branch_labels = None
depends_on = None


def upgrade():

    # transaction
    op.add_column('transaction', sa.Column('patronage', sa.Numeric(precision=9, scale=2), nullable=True))
    op.add_column('transaction', sa.Column('equity_current', sa.Boolean(), nullable=True))
    op.add_column('transaction', sa.Column('self_updated', sa.Boolean(), nullable=True))


def downgrade():

    # transaction
    op.drop_column('transaction', 'self_updated')
    op.drop_column('transaction', 'equity_current')
    op.drop_column('transaction', 'patronage')



# -*- coding: utf-8; -*-
"""add tpr, current price for label batch

Revision ID: d8b0ba4fa795
Revises: 08cc2ef12c18
Create Date: 2022-01-10 13:46:56.056570

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'd8b0ba4fa795'
down_revision = '08cc2ef12c18'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_labels_row
    op.add_column('batch_labels_row', sa.Column('tpr_price', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('batch_labels_row', sa.Column('tpr_starts', sa.DateTime(), nullable=True))
    op.add_column('batch_labels_row', sa.Column('tpr_ends', sa.DateTime(), nullable=True))
    op.add_column('batch_labels_row', sa.Column('current_price', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('batch_labels_row', sa.Column('current_starts', sa.DateTime(), nullable=True))
    op.add_column('batch_labels_row', sa.Column('current_ends', sa.DateTime(), nullable=True))


def downgrade():

    # batch_labels_row
    op.drop_column('batch_labels_row', 'current_ends')
    op.drop_column('batch_labels_row', 'current_starts')
    op.drop_column('batch_labels_row', 'current_price')
    op.drop_column('batch_labels_row', 'tpr_ends')
    op.drop_column('batch_labels_row', 'tpr_starts')
    op.drop_column('batch_labels_row', 'tpr_price')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Mail Monitor Daemon
"""

from __future__ import unicode_literals, absolute_import

import re
import time
import imaplib
# import sys
import logging
# from traceback import format_exception

import six

from rattail.daemon import Daemon
from rattail.mailmon.config import load_mailmon_profiles
from rattail.mailmon.util import (get_lastrun,
                                  get_lastrun_setting,
                                  get_lastrun_timefmt)
from rattail.threads import Thread
from rattail.time import make_utc
from rattail.exceptions import StopProcessing
# from rattail.mail import send_email


log = logging.getLogger(__name__)


class MailMonitorDaemon(Daemon):
    """
    Daemon responsible for checking IMAP folders and detecting email
    messages, and then invoking actions upon them.
    """

    def run(self):
        """
        Starts watcher and worker threads according to configuration.
        """
        monitored = load_mailmon_profiles(self.config)

        # loop thru all accounts
        for key, account in six.iteritems(monitored['__accounts__']):

            # fire up the account watcher thread
            watcher = IMAPAccountWatcher(self.config, account)
            name = 'account_watcher_{}'.format(key)
            log.info("starting IMAP account watcher thread: %s", name)
            thread = Thread(target=watcher, name=name)
            thread.daemon = True
            thread.start()

        # loop indefinitely.  since this is the main thread, the app
        # will terminate when this method ends; all other threads are
        # "subservient" to this one.
        while True:
            time.sleep(.01)


class MailmonConnection(object):
    """
    Abstraction class to hide IMAP details and expose convenience
    methods for use with mailmon daemon.
    """

    def __init__(self, config, account):
        self.config = config
        self.account = account
        self.app = config.get_app()

        self.server = None
        self.server_recycled = None

    def connect(self):

        # do we already have a connection?
        if self.server:

            # keep using forever if no recycle delay
            if not self.account.recycle_delay:
                return

            # okay so does it need to be recycled?
            now = self.app.make_utc()
            if (now - self.server_recycled).seconds < self.account.recycle_delay:

                # nope it's new enough to keep using
                return

            # too old, should recycle
            log.debug("recycle time limit reached (%s seconds), "
                      "disposing of current connection",
                      self.account.recycle_delay)
            self.disconnect()

        # establish new connection
        self.server = imaplib.IMAP4_SSL(self.account.server)
        result = self.server.login(self.account.username, self.account.password)
        log.debug("IMAP server login result: %s", result)
        self.server_recycled = self.app.make_utc()

    def disconnect(self):
        if not self.server:
            return

        try:
            self.server.close()
        except:
            log.debug("server.close() failed!", exc_info=True)
        else:
            try:
                self.server.logout()
            except:
                log.exception("server.logout() failed!")
        finally:
            self.server = None

    def select_folder(self, folder):
        result = self.server.select(folder)
        log.debug("IMAP server select (%s) result: %s", folder, result)
        code, count = result
        if code != 'OK':
            raise RuntimeError("unexpected code '{}' from select() of folder: {}".format(
                code, folder))
        return int(count[0])

    def get_messages(self, folder, criterion):
        self.select_folder(folder)

        # log.debug("invoking IMAP4.search()")
        result = self.server.uid('search', None, criterion)
        try:
            code, items = result
        except:
            log.exception("unexpected search result for folder %s: %s",
                          folder, result)
            raise
        else:
            if code != 'OK':
                raise RuntimeError("IMAP4.search() returned bad code: {}".format(code))
            return items

    def get_all_messages(self, folder):
        return self.get_messages(folder, 'ALL')

    def get_unread_messages(self, folder):
        return self.get_messages(folder, '(UNSEEN)')


class IMAPAccountWatcher(object):
    """
    Abstraction to make watching an IMAP account a little more
    organized.

    Instances of this class are used as callable targets when the
    daemon starts watcher threads.  The instance then is responsible
    for polling the IMAP folder(s) according to the monitor profiles
    associated with the account.
    """

    def __init__(self, config, account):
        self.config = config
        self.account = account
        self.connection = MailmonConnection(config, account)
        self.app = config.get_app()
        self.lastrun_timefmt = get_lastrun_timefmt(self.config)

        # assume all profiles should be monitored at first
        for profile in self.account.monitored:
            profile.monitor = True

    def __call__(self):
        """
        This is the main loop for the account watcher.  It acts as the
        callable target for the thread in which it runs.  It basically
        checks for new messages, processing any found, then waits for
        a spell, then does it again, forever.
        """
        lastruns = {}
        last_error = None
        while True:

            # nb. each profile represents one folder
            for profile in self.account.monitored:

                # we may have stopped monitoring some profiles
                if not profile.monitor:
                    continue

                # track down last run time for this monitor profile
                # (nb. run times are maintained as zone-aware UTC)
                if profile.key in lastruns:
                    lastrun = lastruns[profile.key]
                else:
                    lastrun = get_lastrun(self.config, profile.key)
                    lastruns[profile.key] = lastrun

                # check if we should process this one again yet
                now = self.app.make_utc(tzinfo=True)
                if not lastrun or (now - lastrun).seconds >= profile.imap_delay:

                    # yep, so (try to) process
                    try:
                        self.process_profile(profile)

                    except:
                        log.exception("failed to process profile: %s", profile.key)

                        # presumably the error has to do with our server
                        # connection, so make sure we re-connect next time
                        self.connection.disconnect()

                        # maybe stop monitoring this profile
                        if profile.stop_on_error:
                            profile.monitor = False

                        # skip the rest of this loop; let error delay
                        # handling kick in below
                        last_error = now
                        break

                    else:
                        # poll went okay, so record last run time
                        setting = get_lastrun_setting(self.config, profile.key)
                        session = self.app.make_session()
                        self.app.save_setting(session, setting,
                                              now.strftime(self.lastrun_timefmt))
                        session.commit()
                        session.close()
                        lastruns[profile.key] = now

            # maybe pause for a tick on account of error
            if last_error:
                if self.account.error_delay:
                    time.sleep(self.account.error_delay)
                last_error = None

            # and a general pause just to play nice
            time.sleep(0.01)

    def process_profile(self, profile, unread_only=None):
        """
        Poll a particular IMAP folder and process any messages found.
        """
        self.connection.connect()

        # maybe look for "all" or maybe just "unread"
        if unread_only is None:
            unread_only = profile.imap_unread_only

        if unread_only:
            items = self.connection.get_unread_messages(profile.imap_folder)
        else:
            items = self.connection.get_all_messages(profile.imap_folder)

        # config may dictacte a "max batch size" in which case we will
        # only process so many messages at a time
        uids = items[0].split()
        if profile.max_batch_size:
            if len(uids) > profile.max_batch_size:
                uids = uids[:profile.max_batch_size]

        # process messages
        for uid in uids:
            self.perform_actions(profile, uid)

    def perform_actions(self, profile, msguid):
        """
        Perform all configured actions for the given message uid.
        """
        # ignore if we've stopped monitoring for this profile
        if not profile.monitor:
            return

        # log.debug("queue contained a msguid: %s", msguid)
        for action in profile.actions:
            try:
                self.invoke_action(action, msguid)

            except:
                # stop processing messages altogether for this
                # profile if it is so configured
                if profile.stop_on_error:
                    log.warning("an error was encountered, and config "
                                "dictates that no more actions should be "
                                "processed for profile: %s", profile.key)
                    profile.monitor = False
                    return

                # either way no more actions should be invoked for
                # this particular message
                break

    def invoke_action(self, action, msguid):
        """
        Invoke a single action on a mail message, retrying as necessary.
        """
        attempts = 0
        errtype = None
        while True:
            attempts += 1
            log.debug("invoking action '%s' (attempt #%s of %s) on file: %s",
                      action.spec, attempts, action.retry_attempts, msguid)

            try:
                # TODO: should not reference IMAP server directly?
                action.action(self.connection.server, msguid,
                              *action.args, **action.kwargs)

            except:

                # if we've reached our final attempt, stop retrying
                if attempts >= action.retry_attempts:
                    # log.debug("attempt #%s failed for action '%s' (giving up) on "
                    #           "msguid: %s", attempts, action.spec, msguid,
                    #           exc_info=True)
                    log.exception("attempt #%s failed for action '%s' (giving up) on "
                              "msguid: %s", attempts, action.spec, msguid)
                    # TODO: add email support
                    # exc_type, exc, traceback = sys.exc_info()
                    # send_email(self.config, 'mailmon_action_error', {
                    #     # 'hostname': socket.gethostname(),
                    #     # 'path': path,
                    #     'msguid': msguid,
                    #     'action': action,
                    #     'attempts': attempts,
                    #     'error': exc,
                    #     'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
                    # })
                    raise

                # if this exception is not the first, and is of a
                # different type than seen previously, do *not* continue
                # to retry
                if errtype is not None and not isinstance(error, errtype):
                    log.exception("new exception differs from previous one(s), "
                                  "giving up on action '%s' for msguid: %s",
                                  action.spec, msguid)
                    raise

                # record the type of exception seen, and pause for next retry
                log.warning("attempt #%s failed for action '%s' on msguid: %s",
                            attempts, action.spec, msguid, exc_info=True)
                errtype = type(error)
                log.debug("pausing for %s seconds before making attempt #%s of %s",
                          action.retry_delay, attempts + 1, action.retry_attempts)
                if action.retry_delay:
                    time.sleep(action.retry_delay)

            else:
                # no error, invocation successful
                log.debug("attempt #%s succeeded for action '%s' on msguid: %s",
                          attempts, action.spec, msguid)
                break



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for Generating Projects
"""

from __future__ import unicode_literals, absolute_import

import os
import random
import shutil
import string
import subprocess
import sys

from mako.template import Template


class ProjectHandler(object):
    """
    Base class for project handlers.
    """

    def __init__(self, config):
        self.config = config
        self.app = self.config.get_app()

    def get_all_project_types(self):
        """
        Returns the list of *all* possible project types.
        """
        return [
            'rattail',
            'rattail_integration',
            'tailbone_integration',
            # 'byjove',
            'fabric',
        ]

    def get_supported_project_types(self):
        """
        Returns the list of "supported" project types.
        """
        return self.get_all_project_types()

    def get_storage_dir(self):
        """
        Returns the path to root storage (output) dir for all generated
        projects.
        """
        path = self.config.get('rattail', 'generated_projects.storage_dir')
        if path:
            return path
        return os.path.join(self.config.workdir(require=True),
                            'generated-projects')

    def get_project_dir(self, slug):
        """
        Returns the storage/output path for generated project with the given
        slug name.
        """
        return os.path.join(self.get_storage_dir(), slug)

    def generate_project(self, project_type, slug, options, path=None):
        """
        Generate source code for a new project.

        Note that this method is *not* meant to be overridden by custom
        handlers.  It sticks to the housekeeping chores, and all "interesting"
        logic should be found in :meth:`do_generate()` - which you're free to
        override.

        :param slug: Canonical "slug" (machine-friendly name) for the project.

        :param options: Dictionary(-like) object which contains whatever
           "options" should direct the code generator logic.

        :param path: Path to folder in which project source code should be
           generated.  It will be created if it doesn't exist.  If not
           specified, will use the path returned by :meth:`get_project_dir()`.
        """
        if not path:
            path = self.get_project_dir(slug)

        if os.path.exists(path):
            shutil.rmtree(path)
        os.makedirs(path)

        if project_type == 'rattail_integration':

            if 'name' not in options:
                options['name'] = options['python_project_name']

            if 'slug' not in options:
                options['slug'] = options['name']

            if 'year' not in options:
                options['year'] = self.app.localtime().year

            if 'integration_studly' not in options:
                words = options['integration_name'].split()
                options['integration_studly'] = ''.join([word.capitalize()
                                                         for word in words])

            if 'integration_prefix' not in options:
                words = options['integration_name'].split()
                options['integration_prefix'] = '_'.join([word.lower()
                                                          for word in words])

        if project_type == 'tailbone_integration':

            if 'name' not in options:
                options['name'] = options['python_project_name']

            if 'year' not in options:
                options['year'] = self.app.localtime().year

        # TODO: all of this logic really belongs elsewhere...
        if project_type in ('rattail', 'rattail_integration', 'tailbone_integration', 'fabric'):

            if 'egg_name' not in options:
                options['egg_name'] = options['python_project_name'].replace('-', '_')

            if 'studly_name' not in options:
                words = options['name'].split('-')
                options['studly_name'] = ''.join([word.capitalize()
                                                  for word in words])

            if 'env_name' not in options:
                options['env_name'] = slug

        if project_type == 'rattail':

            if 'db_name' not in options:
                options['db_name'] = slug

            if 'runas_username' not in options:
                options['runas_username'] = slug

            if 'alembic_script_location' not in options:
                if options['extends_db']:
                    location = '{}.db:alembic'.format(options['python_name'])
                else:
                    location = 'rattail.db:alembic'
                options['alembic_script_location'] = location

            if 'alembic_version_locations' not in options:
                locations = ['rattail.db:alembic/versions']
                if options['integrates_catapult']:
                    locations.append('rattail_onager.db:alembic/versions')
                if options['extends_db']:
                    locations.append('{}.db:alembic/versions'.format(options['python_name']))
                options['alembic_version_locations'] = ' '.join(reversed(locations))

            if 'beaker_session_secret' not in options:
                options['beaker_session_secret'] = self.random_string()

        self.do_generate(project_type, slug, options, path)
        return path

    def random_string(self, size=20, chars=string.ascii_letters + string.digits):
        # per https://stackoverflow.com/a/2257449
        return ''.join(random.SystemRandom().choice(chars) for _ in range(size))

    def do_generate(self, project_type, slug, options, path):
        """
        This method supplies the "true" logic for generating new project code.
        Custom handlers may need to override it.  Arguments are essentially the
        same as for :meth:`generate_project()`; however note that the ``path``
        will already exist when this method is invoked.

        Default logic for this method simply runs the ``pcreate`` command with
        the 'rattail' scaffold.  But that is quite limiting, i.e. the only
        "option" it accepts is the project name.  (Hopefully improved logic is
        coming soon though..!)
        """
        pcreate = os.path.join(os.path.dirname(sys.executable), 'pcreate')
        current = os.getcwd()
        os.chdir(path)
        try:
            subprocess.check_call([pcreate, '-s', 'rattail', options['name']])
        finally:
            os.chdir(current)

    def resolve(self, path):
        """
        Returns an absolute path, based on the given path.  So, if that's
        already an absolute path, it's returned as-is; however if not then the
        path is assumed to be relative to this python module.
        """
        abspath = os.path.abspath(path)
        if abspath == path:
            return path
        return os.path.join(os.path.dirname(__file__), path)

    def generate(self, template, output, **context):
        """
        Generate a file from the given template, and save the result to the
        given output path.
        """
        # maybe run it through our simplistic, hand-rolled template engine
        # (note, this is only for the sake of *avoiding* mako logic, when
        # generating "actual" mako templates, so we avoid a mako-within-mako
        # situation.)
        if template.endswith('.mako_tmpl'):
            return self.generate_mako_tmpl(template, output, **context)

        # maybe run it through Mako template engine
        if template.endswith('.mako'):
            return self.generate_mako(template, output, **context)

        # or, just copy the file as-is
        template = self.resolve(template)
        shutil.copyfile(template, output)

    def generate_mako(self, template, output, **context):
        """
        Generate a file from the given template, and save the result to the
        given output path.
        """
        template = self.resolve(template)
        template = Template(filename=template)
        text = template.render(**context)
        with open(output, 'wt') as f:
            f.write(text)

    def generate_mako_tmpl(self, template, output, **context):
        """
        Generate a file from the given template, and save the result to the
        given output path.
        """
        template = os.path.join(os.path.dirname(__file__), template)
        with open(template, 'rt') as f:
            template_lines = f.readlines()

        output_lines = []
        for line in template_lines:
            line = line.rstrip('\n')
            line = line % context
            output_lines.append(line)

        with open(output, 'wt') as f:
            f.write('\n'.join(output_lines))


class RattailProjectHandler(ProjectHandler):
    """
    Project handler for Rattail
    """

    def do_generate(self, project_type, slug, options, path):
        """
        And here we do some experimentation...
        """
        if project_type == 'byjove':
            return self.generate_byjove_project(slug, options, path)
        elif project_type == 'fabric':
            return self.generate_fabric_project(slug, options, path)
        elif project_type == 'rattail_integration':
            return self.generate_rattail_integration_project(slug, options, path)
        elif project_type == 'tailbone_integration':
            return self.generate_tailbone_integration_project(slug, options, path)
        else:
            return self.generate_rattail_project(slug, options, path)

    def generate_rattail_project(self, slug, options, path):
        """
        And here we do some experimentation...
        """
        context = options

        ##############################
        # root project dir
        ##############################

        self.generate('rattail/gitignore.mako', os.path.join(path, '.gitignore'),
                      **context)

        self.generate('rattail/MANIFEST.in.mako', os.path.join(path, 'MANIFEST.in'),
                      **context)

        self.generate('rattail/README.md.mako', os.path.join(path, 'README.md'),
                      **context)

        self.generate('rattail/setup.py.mako', os.path.join(path, 'setup.py'),
                      **context)

        self.generate('rattail/tasks.py.mako', os.path.join(path, 'tasks.py'),
                      **context)

        ##############################
        # root package dir
        ##############################

        package = os.path.join(path, options['python_name'])
        os.makedirs(package)

        self.generate('rattail/package/__init__.py.mako', os.path.join(package, '__init__.py'),
                      **context)

        self.generate('rattail/package/_version.py', os.path.join(package, '_version.py'))

        self.generate('rattail/package/config.py.mako', os.path.join(package, 'config.py'),
                      **context)

        self.generate('rattail/package/commands.py.mako', os.path.join(package, 'commands.py'),
                      **context)

        self.generate('rattail/package/emails.py.mako', os.path.join(package, 'emails.py'),
                      **context)

        self.generate('rattail/package/settings.py', os.path.join(package, 'settings.py'))

        ##############################
        # data dir
        ##############################

        data = os.path.join(package, 'data')
        os.makedirs(data)

        config = os.path.join(data, 'config')
        os.makedirs(config)

        self.generate('rattail/package/data/config/rattail.conf.mako',
                      os.path.join(config, '{}-rattail.conf'.format(slug)),
                      **context)

        self.generate('rattail/package/data/config/web.conf.mako',
                      os.path.join(config, '{}-web.conf'.format(slug)),
                      **context)

        ##############################
        # db package dir
        ##############################

        if context['extends_db']:

            db = os.path.join(package, 'db')
            os.makedirs(db)

            self.generate('rattail/package/db/__init__.py', os.path.join(db, '__init__.py'))

            ####################
            # model
            ####################

            model = os.path.join(db, 'model')
            os.makedirs(model)

            self.generate('rattail/package/db/model/__init__.py.mako', os.path.join(model, '__init__.py'),
                          **context)

            self.generate('rattail/package/db/model/core.py.mako', os.path.join(model, 'core.py'),
                          **context)

            ####################
            # alembic
            ####################

            alembic = os.path.join(db, 'alembic')
            os.makedirs(alembic)

            self.generate('rattail/package/db/alembic/README', os.path.join(alembic, 'README'))

            self.generate('rattail/package/db/alembic/env.py.mako', os.path.join(alembic, 'env.py'),
                          **context)

            self.generate('rattail/package/db/alembic/script.py.mako_', os.path.join(alembic, 'script.py.mako'))

            versions = os.path.join(alembic, 'versions')
            os.makedirs(versions)

            self.generate('rattail/package/db/alembic/versions/.keepme', os.path.join(versions, '.keepme'))

        ##############################
        # web package dir
        ##############################

        if context['has_web']:

            web = os.path.join(package, 'web')
            os.makedirs(web)

            self.generate('rattail/package/web/__init__.py', os.path.join(web, '__init__.py'))

            self.generate('rattail/package/web/app.py.mako', os.path.join(web, 'app.py'),
                          **context)

            self.generate('rattail/package/web/menus.py.mako', os.path.join(web, 'menus.py'),
                          **context)

            self.generate('rattail/package/web/subscribers.py.mako', os.path.join(web, 'subscribers.py'),
                          **context)

            static = os.path.join(web, 'static')
            os.makedirs(static)

            self.generate('rattail/package/web/static/__init__.py.mako', os.path.join(static, '__init__.py'),
                          **context)

            templates = os.path.join(web, 'templates')
            os.makedirs(templates)

            self.generate('rattail/package/web/templates/base_meta.mako_tmpl', os.path.join(templates, 'base_meta.mako'),
                          **context)

            views = os.path.join(web, 'views')
            os.makedirs(views)

            self.generate('rattail/package/web/views/__init__.py.mako', os.path.join(views, '__init__.py'),
                          **context)

            self.generate('rattail/package/web/views/common.py.mako', os.path.join(views, 'common.py'),
                          **context)

        ##############################
        # dev
        ##############################

        dev = os.path.join(path, 'dev')
        os.makedirs(dev)

        self.generate('rattail/dev/bootstrap.py.mako', os.path.join(dev, 'bootstrap.py'),
                      **context)

        self.generate('rattail/dev/tasks.py.mako', os.path.join(dev, 'tasks.py'),
                      **context)

        self.generate('rattail/dev/rattail.conf.mako', os.path.join(dev, 'rattail.conf'),
                      **context)

        self.generate('rattail/dev/web.conf.mako', os.path.join(dev, 'web.conf'),
                      **context)

        ##############################
        # fablib / machines
        ##############################

        if context['uses_fabric']:

            fablib = os.path.join(package, 'fablib')
            os.makedirs(fablib)

            self.generate('rattail/package/fablib/__init__.py.mako', os.path.join(fablib, '__init__.py'),
                          **context)

            self.generate('rattail/package/fablib/python.py.mako', os.path.join(fablib, 'python.py'),
                          **context)

            deploy = os.path.join(fablib, 'deploy')
            os.makedirs(deploy)

            python = os.path.join(deploy, 'python')
            os.makedirs(python)

            self.generate('rattail/package/fablib/deploy/python/premkvirtualenv.mako', os.path.join(python, 'premkvirtualenv.mako'),
                          **context)

            machines = os.path.join(path, 'machines')
            os.makedirs(machines)

            server = os.path.join(machines, 'server')
            os.makedirs(server)

            self.generate('rattail/machines/server/README.md.mako', os.path.join(server, 'README.md'),
                          **context)

            self.generate('rattail/machines/server/Vagrantfile.mako', os.path.join(server, 'Vagrantfile'),
                          **context)

            self.generate('rattail/machines/server/fabenv.py.dist.mako', os.path.join(server, 'fabenv.py.dist'),
                          **context)

            self.generate('rattail/machines/server/fabric.yaml.dist', os.path.join(server, 'fabric.yaml.dist'))

            self.generate('rattail/machines/server/fabfile.py.mako', os.path.join(server, 'fabfile.py'),
                          **context)

            deploy = os.path.join(server, 'deploy')
            os.makedirs(deploy)

            poser = os.path.join(deploy, slug)
            os.makedirs(poser)

            if options['integrates_catapult']:
                self.generate('rattail/machines/server/deploy/poser/freetds.conf.mako_', os.path.join(poser, 'freetds.conf.mako'))
                self.generate('rattail/machines/server/deploy/poser/odbc.ini', os.path.join(poser, 'odbc.ini'))

            self.generate('rattail/machines/server/deploy/poser/rattail.conf.mako', os.path.join(poser, 'rattail.conf.mako'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/cron.conf.mako', os.path.join(poser, 'cron.conf'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/web.conf.mako', os.path.join(poser, 'web.conf.mako'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/supervisor.conf.mako', os.path.join(poser, 'supervisor.conf'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/overnight.sh.mako', os.path.join(poser, 'overnight.sh'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/overnight-wrapper.sh.mako', os.path.join(poser, 'overnight-wrapper.sh'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/crontab.mako', os.path.join(poser, 'crontab.mako'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/upgrade.sh.mako', os.path.join(poser, 'upgrade.sh'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/tasks.py.mako', os.path.join(poser, 'tasks.py'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/upgrade-wrapper.sh.mako', os.path.join(poser, 'upgrade-wrapper.sh'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/sudoers.mako', os.path.join(poser, 'sudoers'),
                          **context)

            self.generate('rattail/machines/server/deploy/poser/logrotate.conf.mako', os.path.join(poser, 'logrotate.conf'),
                          **context)

    def generate_rattail_integration_project(self, slug, options, path):
        """
        And here we do some experimentation...
        """
        context = options

        ##############################
        # root project dir
        ##############################

        self.generate('rattail_integration/gitignore.mako',
                      os.path.join(path, '.gitignore'),
                      **context)

        self.generate('rattail_integration/MANIFEST.in.mako',
                      os.path.join(path, 'MANIFEST.in'),
                      **context)

        self.generate('rattail_integration/README.rst.mako',
                      os.path.join(path, 'README.rst'),
                      **context)

        self.generate('rattail_integration/CHANGELOG.md.mako',
                      os.path.join(path, 'CHANGELOG.md'),
                      **context)

        self.generate('rattail_integration/setup.py.mako',
                      os.path.join(path, 'setup.py'),
                      **context)

        self.generate('rattail_integration/tasks.py.mako',
                      os.path.join(path, 'tasks.py'),
                      **context)

        ##############################
        # root package dir
        ##############################

        package = os.path.join(path, options['python_name'])
        os.makedirs(package)

        self.generate('rattail_integration/package/__init__.py.mako',
                      os.path.join(package, '__init__.py'),
                      **context)

        self.generate('rattail_integration/package/_version.py',
                      os.path.join(package, '_version.py'))

        if context['extends_config']:
            self.generate('rattail_integration/package/config.py.mako',
                          os.path.join(package, 'config.py'),
                          **context)

        ##############################
        # db package dir
        ##############################

        if context['extends_db']:

            db = os.path.join(package, 'db')
            os.makedirs(db)

            self.generate('rattail_integration/package/db/__init__.py',
                          os.path.join(db, '__init__.py'))

            ####################
            # model
            ####################

            model = os.path.join(db, 'model')
            os.makedirs(model)

            self.generate('rattail_integration/package/db/model/__init__.py.mako',
                          os.path.join(model, '__init__.py'),
                          **context)

            self.generate('rattail_integration/package/db/model/customers.py.mako',
                          os.path.join(model, 'customers.py'),
                          **context)

            ####################
            # alembic
            ####################

            alembic = os.path.join(db, 'alembic')
            os.makedirs(alembic)

            versions = os.path.join(alembic, 'versions')
            os.makedirs(versions)

            self.generate('rattail_integration/package/db/alembic/versions/.keepme',
                          os.path.join(versions, '.keepme'))

    def generate_tailbone_integration_project(self, slug, options, path):
        """
        And here we do some experimentation...
        """
        context = options

        ##############################
        # root project dir
        ##############################

        self.generate('tailbone_integration/gitignore.mako',
                      os.path.join(path, '.gitignore'),
                      **context)

        self.generate('tailbone_integration/MANIFEST.in.mako',
                      os.path.join(path, 'MANIFEST.in'),
                      **context)

        self.generate('tailbone_integration/README.rst.mako',
                      os.path.join(path, 'README.rst'),
                      **context)

        self.generate('tailbone_integration/CHANGELOG.md.mako',
                      os.path.join(path, 'CHANGELOG.md'),
                      **context)

        self.generate('tailbone_integration/setup.py.mako',
                      os.path.join(path, 'setup.py'),
                      **context)

        self.generate('tailbone_integration/tasks.py.mako',
                      os.path.join(path, 'tasks.py'),
                      **context)

        ##############################
        # root package dir
        ##############################

        package = os.path.join(path, options['python_name'])
        os.makedirs(package)

        self.generate('tailbone_integration/package/__init__.py.mako',
                      os.path.join(package, '__init__.py'),
                      **context)

        self.generate('tailbone_integration/package/_version.py',
                      os.path.join(package, '_version.py'))

        ##############################
        # views
        ##############################

        views = os.path.join(package, 'views')
        os.makedirs(views)

        self.generate('tailbone_integration/package/views/__init__.py',
                      os.path.join(views, '__init__.py'))

        ##############################
        # static
        ##############################

        static = os.path.join(package, 'static')
        os.makedirs(static)

        self.generate('tailbone_integration/package/static/__init__.py.mako',
                      os.path.join(static, '__init__.py'),
                      **context)

        ##############################
        # templates
        ##############################

        templates = os.path.join(package, 'templates')
        os.makedirs(templates)

        self.generate('tailbone_integration/package/templates/.keepme',
                      os.path.join(templates, '.keepme'))

    def generate_byjove_project(self, slug, options, path):
        """
        Generate a new 'byjove' project per the given arguments.
        """
        context = options

        ##############################
        # root project dir
        ##############################

        self.generate('byjove/CHANGELOG.md.mako', os.path.join(path, 'CHANGELOG.md'),
                      **context)

        self.generate('byjove/gitignore', os.path.join(path, '.gitignore'))

        self.generate('byjove/README.md.mako', os.path.join(path, 'README.md'),
                      **context)

        self.generate('byjove/vue.config.js.dist.mako', os.path.join(path, 'vue.config.js.dist'),
                      **context)

    def generate_fabric_project(self, slug, options, path):
        """
        Generate a new 'fabric' project per the given arguments.
        """
        context = options

        ##############################
        # root project dir
        ##############################

        self.generate('fabric/gitignore.mako', os.path.join(path, '.gitignore'),
                      **context)

        self.generate('fabric/README.md.mako', os.path.join(path, 'README.md'),
                      **context)

        self.generate('fabric/setup.py.mako', os.path.join(path, 'setup.py'),
                      **context)

        ##############################
        # package dir
        ##############################

        package = os.path.join(path, options['python_name'])
        os.makedirs(package)

        self.generate('fabric/package/__init__.py.mako', os.path.join(package, '__init__.py'),
                      **context)

        self.generate('fabric/package/_version.py', os.path.join(package, '_version.py'))

        ##############################
        # machines
        ##############################

        machines = os.path.join(path, 'machines')
        os.makedirs(machines)

        ##############################
        # generic-server
        ##############################

        generic_server = os.path.join(machines, 'generic-server')
        os.makedirs(generic_server)

        self.generate('fabric/machines/generic-server/README.md.mako', os.path.join(generic_server, 'README.md'),
                      **context)

        self.generate('fabric/machines/generic-server/Vagrantfile.mako', os.path.join(generic_server, 'Vagrantfile'),
                      **context)

        self.generate('fabric/machines/generic-server/fabenv.py.dist.mako', os.path.join(generic_server, 'fabenv.py.dist'),
                      **context)

        self.generate('fabric/machines/generic-server/fabric.yaml.dist', os.path.join(generic_server, 'fabric.yaml.dist'))

        self.generate('fabric/machines/generic-server/fabfile.py.mako', os.path.join(generic_server, 'fabfile.py'),
                      **context)

        ##############################
        # theo-server
        ##############################

        theo_server = os.path.join(machines, 'theo-server')
        os.makedirs(theo_server)

        self.generate('fabric/machines/theo-server/README.md', os.path.join(theo_server, 'README.md'))

        self.generate('fabric/machines/theo-server/Vagrantfile', os.path.join(theo_server, 'Vagrantfile'))

        self.generate('fabric/machines/theo-server/fabenv.py.dist.mako', os.path.join(theo_server, 'fabenv.py.dist'),
                      **context)

        self.generate('fabric/machines/theo-server/fabric.yaml.dist', os.path.join(theo_server, 'fabric.yaml.dist'))

        self.generate('fabric/machines/theo-server/fabfile.py.mako', os.path.join(theo_server, 'fabfile.py'),
                      **context)

        theo_deploy = os.path.join(theo_server, 'deploy')
        os.makedirs(theo_deploy)

        theo_python = os.path.join(theo_deploy, 'python')
        os.makedirs(theo_python)

        self.generate('fabric/machines/theo-server/deploy/python/pip.conf.mako', os.path.join(theo_python, 'pip.conf.mako'),
                      **context)

        theo_rattail = os.path.join(theo_deploy, 'rattail')
        os.makedirs(theo_rattail)

        self.generate('fabric/machines/theo-server/deploy/rattail/rattail.conf.mako', os.path.join(theo_rattail, 'rattail.conf.mako'),
                      **context)

        self.generate('fabric/machines/theo-server/deploy/rattail/freetds.conf.mako_', os.path.join(theo_rattail, 'freetds.conf.mako'))

        self.generate('fabric/machines/theo-server/deploy/rattail/odbc.ini', os.path.join(theo_rattail, 'odbc.ini'))

        theo_theo_common = os.path.join(theo_deploy, 'theo-common')
        os.makedirs(theo_theo_common)

        self.generate('fabric/machines/theo-server/deploy/theo-common/rattail.conf.mako', os.path.join(theo_theo_common, 'rattail.conf.mako'),
                      **context)

        self.generate('fabric/machines/theo-server/deploy/theo-common/web.conf.mako', os.path.join(theo_theo_common, 'web.conf.mako'),
                      **context)

        self.generate('fabric/machines/theo-server/deploy/theo-common/upgrade.sh.mako', os.path.join(theo_theo_common, 'upgrade.sh.mako'),
                      **context)

        self.generate('fabric/machines/theo-server/deploy/theo-common/tasks.py.mako_', os.path.join(theo_theo_common, 'tasks.py.mako'))

        self.generate('fabric/machines/theo-server/deploy/theo-common/upgrade-wrapper.sh.mako_', os.path.join(theo_theo_common, 'upgrade-wrapper.sh.mako'))

        self.generate('fabric/machines/theo-server/deploy/theo-common/supervisor.conf.mako_', os.path.join(theo_theo_common, 'supervisor.conf.mako'))

        self.generate('fabric/machines/theo-server/deploy/theo-common/sudoers.mako_', os.path.join(theo_theo_common, 'sudoers.mako'))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail data import
"""

from __future__ import unicode_literals, absolute_import

import logging

import sqlalchemy as sa

from rattail.db import Session
from rattail.importing import model
from rattail.importing.handlers import FromSQLAlchemyHandler, ToSQLAlchemyHandler
from rattail.importing.sqlalchemy import FromSQLAlchemySameToSame
from rattail.util import OrderedDict


log = logging.getLogger(__name__)


class FromRattailHandler(FromSQLAlchemyHandler):
    """
    Base class for import handlers which target a Rattail database on the local side.
    """
    host_key = 'rattail'
    generic_host_title = "Rattail"

    @property
    def host_title(self):
        return self.config.app_title(default="Rattail")

    def make_host_session(self):
        return Session()


class ToRattailHandler(ToSQLAlchemyHandler):
    """
    Base class for import handlers which target a Rattail database on the local side.
    """
    generic_local_title = "Rattail"
    local_key = 'rattail'

    @property
    def local_title(self):
        return self.config.app_title(default="Rattail")

    def make_session(self):
        kwargs = {}
        if hasattr(self, 'runas_user'):
            kwargs['continuum_user'] = self.runas_user
        return Session(**kwargs)

    def begin_local_transaction(self):
        self.session = self.make_session()

        # load "runas user" into current session
        if hasattr(self, 'runas_user') and self.runas_user:
            dbmodel = self.config.get_model()
            runas_user = self.session.query(dbmodel.User)\
                                     .get(self.runas_user.uuid)
            if not runas_user:
                log.info("runas_user does not exist in target session: %s",
                         self.runas_user.username)
            # this may be None if user does not exist in target session
            self.runas_user = runas_user

        # declare "runas user" is data versioning author
        if hasattr(self, 'runas_username') and self.runas_username:
            self.session.set_continuum_user(self.runas_username)


class FromRattailToRattailBase(object):
    """
    Common base class for Rattail -> Rattail data import/export handlers.
    """

    def get_importers(self):
        importers = OrderedDict()
        importers['Person'] = PersonImporter
        importers['GlobalPerson'] = GlobalPersonImporter
        importers['PersonEmailAddress'] = PersonEmailAddressImporter
        importers['PersonPhoneNumber'] = PersonPhoneNumberImporter
        importers['PersonMailingAddress'] = PersonMailingAddressImporter
        importers['MergePeopleRequest'] = MergePeopleRequestImporter
        importers['Role'] = RoleImporter
        importers['GlobalRole'] = GlobalRoleImporter
        importers['User'] = UserImporter
        importers['AdminUser'] = AdminUserImporter
        importers['GlobalUser'] = GlobalUserImporter
        importers['Message'] = MessageImporter
        importers['MessageRecipient'] = MessageRecipientImporter
        importers['Store'] = StoreImporter
        importers['StorePhoneNumber'] = StorePhoneNumberImporter
        importers['Employee'] = EmployeeImporter
        importers['EmployeeStore'] = EmployeeStoreImporter
        importers['EmployeeEmailAddress'] = EmployeeEmailAddressImporter
        importers['EmployeePhoneNumber'] = EmployeePhoneNumberImporter
        importers['ScheduledShift'] = ScheduledShiftImporter
        importers['WorkedShift'] = WorkedShiftImporter
        importers['Customer'] = CustomerImporter
        importers['CustomerGroup'] = CustomerGroupImporter
        importers['CustomerGroupAssignment'] = CustomerGroupAssignmentImporter
        importers['CustomerPerson'] = CustomerPersonImporter
        importers['CustomerEmailAddress'] = CustomerEmailAddressImporter
        importers['CustomerPhoneNumber'] = CustomerPhoneNumberImporter
        importers['Member'] = MemberImporter
        importers['MemberEmailAddress'] = MemberEmailAddressImporter
        importers['MemberPhoneNumber'] = MemberPhoneNumberImporter
        importers['Vendor'] = VendorImporter
        importers['VendorEmailAddress'] = VendorEmailAddressImporter
        importers['VendorPhoneNumber'] = VendorPhoneNumberImporter
        importers['VendorContact'] = VendorContactImporter
        importers['Department'] = DepartmentImporter
        importers['EmployeeDepartment'] = EmployeeDepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Category'] = CategoryImporter
        importers['Family'] = FamilyImporter
        importers['ReportCode'] = ReportCodeImporter
        importers['DepositLink'] = DepositLinkImporter
        importers['Tax'] = TaxImporter
        importers['InventoryAdjustmentReason'] = InventoryAdjustmentReasonImporter
        importers['Brand'] = BrandImporter
        importers['Product'] = ProductImporter
        importers['ProductCode'] = ProductCodeImporter
        importers['ProductCost'] = ProductCostImporter
        importers['ProductPrice'] = ProductPriceImporter
        importers['ProductPriceAssociation'] = ProductPriceAssociationImporter
        importers['ProductStoreInfo'] = ProductStoreInfoImporter
        importers['ProductVolatile'] = ProductVolatileImporter
        importers['ProductImage'] = ProductImageImporter
        importers['LabelProfile'] = LabelProfileImporter
        return importers

    def get_default_keys(self):
        keys = self.get_importer_keys()

        avoid_by_default = [
            'Role',
            'GlobalRole',
            'GlobalPerson',
            'AdminUser',
            'GlobalUser',
            'ProductImage',
            'ProductPriceAssociation',
        ]

        for key in avoid_by_default:
            if key in keys:
                keys.remove(key)

        return keys


class FromRattailToRattailImport(FromRattailToRattailBase, FromRattailHandler, ToRattailHandler):
    """
    Handler for Rattail (other) -> Rattail (local) data import.

    .. attribute:: direction

       Value is ``'import'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    dbkey = 'other'

    @property
    def host_title(self):
        return "{} ({})".format(self.config.app_title(default="Rattail"), self.dbkey)

    @property
    def local_title(self):
        return self.config.node_title(default="Rattail (local)")

    def make_host_session(self):
        return Session(bind=self.config.rattail_engines[self.dbkey])


class FromRattailToRattailExport(FromRattailToRattailBase, FromRattailHandler, ToRattailHandler):
    """
    Handler for Rattail (local) -> Rattail (other) data export.

    .. attribute:: direction

       Value is ``'export'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    direction = 'export'
    dbkey = 'other'

    @property
    def host_title(self):
        return self.config.node_title()

    @property
    def local_title(self):
        return "{} ({})".format(self.config.app_title(default="Rattail"), self.dbkey)

    def make_session(self):
        return Session(bind=self.config.rattail_engines[self.dbkey])


class FromRattail(FromSQLAlchemySameToSame):
    """
    Base class for Rattail -> Rattail data importers.
    """


class PersonImporter(FromRattail, model.PersonImporter):
    pass


class GlobalPersonImporter(FromRattail, model.GlobalPersonImporter):
    """
    This is a customized version of the :class:`PersonImporter`, which simply
    avoids "local only" person accounts.
    """

    def query(self):
        query = super(GlobalPersonImporter, self).query()

        # never include "local only" people
        query = query.filter(sa.or_(
            self.host_model_class.local_only == False,
            self.host_model_class.local_only == None))

        return query

    def normalize_host_object(self, person):

        # must check this here for sake of datasync
        if person.local_only:
            return

        data = super(GlobalPersonImporter, self).normalize_host_object(person)
        return data


class PersonEmailAddressImporter(FromRattail, model.PersonEmailAddressImporter):
    pass

class PersonPhoneNumberImporter(FromRattail, model.PersonPhoneNumberImporter):
    pass

class PersonMailingAddressImporter(FromRattail, model.PersonMailingAddressImporter):
    pass

class MergePeopleRequestImporter(FromRattail, model.MergePeopleRequestImporter):
    pass

class RoleImporter(FromRattail, model.RoleImporter):
    pass


class GlobalRoleImporter(RoleImporter):
    """
    Role importer which only will handle roles which have the
    :attr:`~rattail.db.model.users.Role.sync_me` flag set.  (So it
    syncs those roles but avoids others.)
    """

    @property
    def supported_fields(self):
        fields = list(super(GlobalRoleImporter, self).supported_fields)
        fields.extend([
            'permissions',
            'users',
        ])
        return fields

    # nb. we must override both cache_query() and query() b/c they use
    # different sessions!

    def cache_query(self):
        """
        Return the query to be used when caching "local" data.
        """
        query = super(GlobalRoleImporter, self).cache_query()
        model = self.model

        # only want roles which are *meant* to be synced
        query = query.filter(model.Role.sync_me == True)

        return query

    def query(self):
        query = super(GlobalRoleImporter, self).query()
        model = self.model

        # only want roles which are *meant* to be synced
        query = query.filter(model.Role.sync_me == True)

        return query

    # nb. we do not need to override normalize_host_object() b/c it
    # just calls normalize_local_object() by default

    def normalize_local_object(self, role):

        # only want roles which are *meant* to be synced
        if not role.sync_me:
            return

        data = super(GlobalRoleImporter, self).normalize_local_object(role)
        if data:

            # users
            if 'users' in self.fields:
                data['users'] = sorted([user.uuid for user in role.users])

            # permissions
            if 'permissions' in self.fields:
                auth = self.app.get_auth_handler()
                perms = auth.cache_permissions(self.session, role,
                                               include_guest=False)
                data['permissions'] = sorted(perms)

            return data

    def update_object(self, role, host_data, local_data=None, **kwargs):
        role = super(GlobalRoleImporter, self).update_object(role, host_data,
                                                             local_data=local_data,
                                                             **kwargs)
        model = self.model

        # users
        # nb. we only update users if this role has flag set
        if 'users' in self.fields and role.sync_users:

            new_users = host_data['users']
            old_users = local_data['users'] if local_data else []
            changed = False

            # add some users
            for new_user in new_users:
                if new_user not in old_users:
                    user = self.session.query(model.User).get(new_user)
                    if user:
                        user.roles.append(role)
                        changed = True

            # remove some users
            for old_user in old_users:
                if old_user not in new_users:
                    user = self.session.query(model.User).get(old_user)
                    if user:
                        user.roles.remove(role)
                        changed = True

            if changed:
                self.session.flush()
                self.session.refresh(role)
                # also record a change to the role, for datasync.
                # this is done "just in case" the role is to be
                # synced to all nodes
                if self.session.rattail_record_changes:
                    self.session.add(model.Change(class_name='Role',
                                                  instance_uuid=role.uuid,
                                                  deleted=False))

        # permissions
        if 'permissions' in self.fields:
            auth = self.app.get_auth_handler()
            new_perms = host_data['permissions']
            old_perms = local_data['permissions'] if local_data else []

            # grant permissions
            for new_perm in new_perms:
                if new_perm not in old_perms:
                    auth.grant_permission(role, new_perm)

            # revoke permissions
            for old_perm in old_perms:
                if old_perm not in new_perms:
                    auth.revoke_permission(role, old_perm)

        return role


class UserImporter(FromRattail, model.UserImporter):
    pass


class GlobalUserImporter(FromRattail, model.GlobalUserImporter):
    """
    This is a customized version of the :class:`UserImporter`, which simply
    avoids "local only" user accounts.
    """

    def query(self):
        query = super(GlobalUserImporter, self).query()

        # never include "local only" users
        query = query.filter(sa.or_(
            self.host_model_class.local_only == False,
            self.host_model_class.local_only == None))

        return query

    def normalize_host_object(self, user):

        # must check this here for sake of datasync
        if user.local_only:
            return

        data = super(GlobalUserImporter, self).normalize_host_object(user)
        return data


class AdminUserImporter(FromRattail, model.AdminUserImporter):

    @property
    def supported_fields(self):
        return super(AdminUserImporter, self).supported_fields + [
            'admin',
        ]

    def normalize_host_object(self, user):
        data = super(AdminUserImporter, self).normalize_local_object(user) # sic
        if 'admin' in self.fields: # TODO: do we really need this, after the above?
            data['admin'] = self.admin_uuid in [r.role_uuid for r in user._roles]
        return data


class MessageImporter(FromRattail, model.MessageImporter):
    pass

class MessageRecipientImporter(FromRattail, model.MessageRecipientImporter):
    pass

class StoreImporter(FromRattail, model.StoreImporter):
    pass

class StorePhoneNumberImporter(FromRattail, model.StorePhoneNumberImporter):
    pass

class EmployeeImporter(FromRattail, model.EmployeeImporter):
    pass

class EmployeeStoreImporter(FromRattail, model.EmployeeStoreImporter):
    pass

class EmployeeDepartmentImporter(FromRattail, model.EmployeeDepartmentImporter):
    pass

class EmployeeEmailAddressImporter(FromRattail, model.EmployeeEmailAddressImporter):
    pass

class EmployeePhoneNumberImporter(FromRattail, model.EmployeePhoneNumberImporter):
    pass

class ScheduledShiftImporter(FromRattail, model.ScheduledShiftImporter):
    pass

class WorkedShiftImporter(FromRattail, model.WorkedShiftImporter):
    pass

class CustomerImporter(FromRattail, model.CustomerImporter):
    pass

class CustomerGroupImporter(FromRattail, model.CustomerGroupImporter):
    pass

class CustomerGroupAssignmentImporter(FromRattail, model.CustomerGroupAssignmentImporter):
    pass

class CustomerPersonImporter(FromRattail, model.CustomerPersonImporter):
    pass

class CustomerEmailAddressImporter(FromRattail, model.CustomerEmailAddressImporter):
    pass

class CustomerPhoneNumberImporter(FromRattail, model.CustomerPhoneNumberImporter):
    pass

class MemberImporter(FromRattail, model.MemberImporter):
    pass

class MemberEmailAddressImporter(FromRattail, model.MemberEmailAddressImporter):
    pass

class MemberPhoneNumberImporter(FromRattail, model.MemberPhoneNumberImporter):
    pass

class VendorImporter(FromRattail, model.VendorImporter):
    pass

class VendorEmailAddressImporter(FromRattail, model.VendorEmailAddressImporter):
    pass

class VendorPhoneNumberImporter(FromRattail, model.VendorPhoneNumberImporter):
    pass

class VendorContactImporter(FromRattail, model.VendorContactImporter):
    pass

class DepartmentImporter(FromRattail, model.DepartmentImporter):
    pass

class SubdepartmentImporter(FromRattail, model.SubdepartmentImporter):
    pass

class CategoryImporter(FromRattail, model.CategoryImporter):
    pass

class FamilyImporter(FromRattail, model.FamilyImporter):
    pass

class ReportCodeImporter(FromRattail, model.ReportCodeImporter):
    pass

class DepositLinkImporter(FromRattail, model.DepositLinkImporter):
    pass

class TaxImporter(FromRattail, model.TaxImporter):
    pass

class InventoryAdjustmentReasonImporter(FromRattail, model.InventoryAdjustmentReasonImporter):
    pass

class BrandImporter(FromRattail, model.BrandImporter):
    pass


class ProductWithPriceImporter(FromRattail, model.ProductImporter):
    """
    This can perhaps be thought of as the "complete" Product record
    importer.  The "normal" Product importer will typically avoid the
    "price uuid" reference fields, b/c of that foreign key chaos.

    Note that this importer is not (yet?) used directly, but is
    primarily useful as a base class.
    """
    # these require special handling due to the 2-way table dependency
    price_reference_fields = [
        'regular_price_uuid',
        'tpr_price_uuid',
        'sale_price_uuid',
        'current_price_uuid',
        'suggested_price_uuid',
    ]

    def query(self):
        query = super(ProductWithPriceImporter, self).query()

        # make sure potential unit items (i.e. rows with NULL unit_uuid) come
        # first, so they will be created before pack items reference them
        # cf. https://www.postgresql.org/docs/current/static/queries-order.html
        # cf. https://stackoverflow.com/a/7622046
        query = query.order_by(self.host_model_class.unit_uuid.desc())

        return query


class ProductPriceAssociationImporter(ProductWithPriceImporter):
    """
    Note that this importer is *only* for sake of handling the "price
    uuid" fields.
    """

    @property
    def simple_fields(self):
        return ['uuid'] + self.price_reference_fields


class ProductImporter(ProductWithPriceImporter):
    """
    Note that this is the "normal" Product record importer, but it
    inherits from the "complete" importer.  This one avoids the "price
    uuid" fields to avoid that foreign key chaos.
    """

    @property
    def simple_fields(self):
        fields = super(ProductImporter, self).simple_fields
        # NOTE: it seems we can't consider these "simple" due to the
        # self-referencing foreign key situation.  an importer can still
        # "support" these fields, but they're excluded from the simple set for
        # sake of rattail <-> rattail
        for field in self.price_reference_fields:
            fields.remove(field)
        return fields


class ProductCodeImporter(FromRattail, model.ProductCodeImporter):
    pass

class ProductCostImporter(FromRattail, model.ProductCostImporter):
    pass

class ProductPriceImporter(FromRattail, model.ProductPriceImporter):

    @property
    def supported_fields(self):
        # nb. parent class FromRattail only supports simple_fields, so
        # we explicitly copy logic from model importer class here.
        return self.simple_fields + self.product_reference_fields


class ProductStoreInfoImporter(FromRattail, model.ProductStoreInfoImporter):
    pass

class ProductVolatileImporter(FromRattail, model.ProductVolatileImporter):
    pass


class ProductImageImporter(FromRattail, model.ProductImageImporter):
    """
    Importer for product images.  Note that this uses the "batch" approach
    because fetching all data up front is not performant when the host/local
    systems are on different machines etc.
    """

    def query(self):
        query = self.host_session.query(self.model_class)\
                                 .order_by(self.model_class.uuid)
        return query[self.host_index:self.host_index + self.batch_size]


class LabelProfileImporter(FromRattail, model.LabelProfileImporter):

    def query(self):
        query = super(LabelProfileImporter, self).query()

        if not self.config.getbool('rattail', 'labels.sync_all_profiles', default=False):
            # only fetch labels from host which are marked as "sync me"
            query = query .filter(self.model_class.sync_me == True)

        return query.order_by(self.model_class.ordinal)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Pricing Utilities
"""

from __future__ import unicode_literals, absolute_import


def gross_margin(price, cost, percentage=False):
    """
    Calculate and return a gross margin percentage based on ``price`` and
    ``cost``.

    Please note, that for historical reasons, the default behavior is to return
    the margin as a decimal value from 0.0 through 100.0 (or beyond, perhaps).

    However the "industry standard" seems to be to use a decimal value between
    0.000 and 1.000 instead.  Specify ``percentage=True`` for this behavior.

    If ``price`` is empty (or zero), returns ``None``.

    If ``cost`` is empty (or zero), returns ``100`` (or ``1`` if
    ``percentage=True``).
    """
    if not price:
        return None

    if not cost:
        if percentage:
            return 1
        return 100

    margin = (price - cost) / price
    if percentage:
        return margin
    return 100 * margin


def calculate_markup(margin):
    """
    Calculate the "markup" value corresponding to the given margin.

    :param margin: Profit margin as decimal percentage (e.g. between
       0.0 and 1.0).

    :returns: Equivalent cost markup as decimal value (e.g. 1.4).
    """
    if margin is None:
        return
    if margin == 0:
        return 1

    return 1 / (1 - margin)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Windows Services
"""

from __future__ import unicode_literals

import sys
import subprocess
import logging

try:
    import win32serviceutil
except ImportError:
    # Mock out for Linux.
    class Object(object):
        pass
    win32serviceutil = Object()
    win32serviceutil.ServiceFramework = Object

from rattail.config import make_config


log = logging.getLogger(__name__)


class Service(win32serviceutil.ServiceFramework):
    """
    Base class for Windows service implementations.
    """

    def __init__(self, args):
        """
        Constructor.
        """

        import win32event

        win32serviceutil.ServiceFramework.__init__(self, args)

        # Create "wait stop" event, for main worker loop.
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)

    def Initialize(self, config):
        """
        Service initialization.
        """
        return True

    def SvcDoRun(self):
        """
        This method is invoked when the service starts.
        """

        import win32service
        import win32event
        import servicemanager
        from pywintypes import error
        from winerror import ERROR_LOG_FILE_FULL

        # Write start occurrence to Windows Event Log.
        try:
            servicemanager.LogMsg(
                servicemanager.EVENTLOG_INFORMATION_TYPE,
                servicemanager.PYS_SERVICE_STARTED,
                (self._svc_name_, ''))
        except error as e:
            if e.winerror == ERROR_LOG_FILE_FULL:
                log.error("SvcDoRun: Windows event log is full!")
            else:
                raise

        # Read configuration file(s).
        logging.basicConfig()
        config = make_config(winsvc=self._svc_name_)
        config.configure_logging()

        # Figure out what we're supposed to be doing.
        if self.Initialize(config):

            # Wait infinitely for stop request, while threads do their thing.
            log.info("SvcDoRun: All threads started; waiting for stop request.")
            win32event.WaitForSingleObject(self.hWaitStop, win32event.INFINITE)
            log.info("SvcDoRun: Stop request received.")

        else: # Nothing to be done...
            msg = "Nothing to do!  (Initialization failed.)"
            log.warning("SvcDoRun: {0}".format(msg))
            servicemanager.LogWarningMsg(msg)
            self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)

        # Write stop occurrence to Windows Event Log.
        try:
            servicemanager.LogMsg(
                servicemanager.EVENTLOG_INFORMATION_TYPE,
                servicemanager.PYS_SERVICE_STOPPED,
                (self._svc_name_, ''))
        except error as e:
            if e.winerror == ERROR_LOG_FILE_FULL:
                log.error("SvcDoRun: Windows event log is full!")
            else:
                raise

    def SvcStop(self):
        """
        This method is invoked when the service is requested to stop itself.
        """

        import win32service
        import win32event

        # Let the SCM know we're trying to stop.
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)

        # Let worker loop know its job is done.
        win32event.SetEvent(self.hWaitStop)


def execute_service_command(module, command, *args):
    """
    Execute a command for a Windows service.

    :param module: A proper Python module object, which must implement a
       command line interface when invoked directly by the Python interpreter.
       This interface is assumed to be implemented via the function
       ``win32serviceutil.HandleCommandLine()``.

    :param command: The name of a command supported by the service interface.
       This is presumably one of:

       * ``'install'``
       * ``'remove'``
       * ``'start'``
       * ``'stop'``
       * ``'restart'``

    :param \*args: If any additional arguments are present, they are assumed to
       be "option" arguments and will precede ``command`` when the actual
       command line is constructed.
    """
    command = [command]
    if args:
        command = list(args) + command
    subprocess.call([sys.executable, module.__file__] + command)


def delayed_auto_start_service(name):
    """
    Configure a Windows service for delayed automatic startup.

    :param name: The system-recognized name of the service (i.e. not the
       display name).

    .. note::
       It is assumed that the service is already configured to start
       automatically.  This function only modifies the service so that its
       automatic startup is delayed.
    """
    import win32service

    hSCM = win32service.OpenSCManager(
        None,
        None,
        win32service.SC_MANAGER_ENUMERATE_SERVICE)

    hService = win32service.OpenService(
        hSCM,
        name,
        win32service.SERVICE_CHANGE_CONFIG)

    win32service.ChangeServiceConfig2(
        hService,
        win32service.SERVICE_CONFIG_DELAYED_AUTO_START_INFO,
        True)

    win32service.CloseServiceHandle(hService)
    win32service.CloseServiceHandle(hSCM)



# -*- coding: utf-8 -*-
"""remove old versioning

Revision ID: e9fca3d787ea
Revises: a5a7358ed27f
Create Date: 2017-05-25 17:38:12.775386

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'e9fca3d787ea'
down_revision = u'a5a7358ed27f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # *_version
    op.drop_table('label_profile_version')
    op.drop_table('customer_group_version')
    op.drop_table('product_code_version')
    op.drop_table('store_version')
    op.drop_table('address_version')
    op.drop_table('employee_version')
    op.drop_table('email_version')
    op.drop_table('customer_version')
    op.drop_table('user_version')
    op.drop_table('product_price_version')
    op.drop_table('person_version')
    op.drop_table('brand_version')
    op.drop_table('tax_version')
    op.drop_table('subdepartment_version')
    op.drop_table('department_version')
    op.drop_table('deposit_link_version')
    op.drop_table('family_version')
    op.drop_table('report_code_version')
    op.drop_table('vendor_version')
    op.drop_table('phone_version')
    op.drop_table('product_cost_version')
    op.drop_table('category_version')
    op.drop_table('role_version')
    op.drop_table('product_version')

    # transaction
    op.drop_table('transaction')


def downgrade():

    # transaction
    op.create_table('transaction',
                    sa.Column('issued_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
                    sa.Column('id', sa.BIGINT(), nullable=False),
                    sa.Column('remote_addr', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('user_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.ForeignKeyConstraint(['user_uuid'], [u'user.uuid'], name=u'transaction_user_uuid_fkey'),
                    sa.PrimaryKeyConstraint('id', name=u'transaction_pkey')
    )

    # product_version
    op.create_table('product_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('upc', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('subdepartment_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('category_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('family_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('report_code_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('brand_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('description', sa.VARCHAR(length=60), autoincrement=False, nullable=True),
                    sa.Column('description2', sa.VARCHAR(length=60), autoincrement=False, nullable=True),
                    sa.Column('size', sa.VARCHAR(length=30), autoincrement=False, nullable=True),
                    sa.Column('unit_of_measure', sa.VARCHAR(length=4), autoincrement=False, nullable=True),
                    sa.Column('not_for_sale', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.Column('regular_price_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('current_price_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.Column('deleted', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.Column('case_pack', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('organic', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.Column('deposit_link_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('tax_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('unit_size', sa.NUMERIC(precision=8, scale=3), autoincrement=False, nullable=True),
                    sa.Column('weighed', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.Column('discountable', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.Column('special_order', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.Column('last_sold', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'product_version_pkey')
    )

    # role_version
    op.create_table('role_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('name', sa.VARCHAR(length=25), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'role_version_pkey')
    )

    # category_version
    op.create_table('category_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('number', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'category_version_pkey')
    )

    # product_cost_version
    op.create_table('product_cost_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('product_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('vendor_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
                    sa.Column('case_size', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('case_cost', sa.NUMERIC(precision=9, scale=5), autoincrement=False, nullable=True),
                    sa.Column('pack_size', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('pack_cost', sa.NUMERIC(precision=9, scale=5), autoincrement=False, nullable=True),
                    sa.Column('unit_cost', sa.NUMERIC(precision=9, scale=5), autoincrement=False, nullable=True),
                    sa.Column('effective', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'product_cost_version_pkey')
    )

    # phone_version
    op.create_table('phone_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('type', sa.VARCHAR(length=15), autoincrement=False, nullable=True),
                    sa.Column('number', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'phone_version_pkey')
    )

    # vendor_version
    op.create_table('vendor_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.VARCHAR(length=15), autoincrement=False, nullable=True),
                    sa.Column('name', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('special_discount', sa.NUMERIC(precision=5, scale=3), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.Column('lead_time_days', sa.NUMERIC(precision=5, scale=1), autoincrement=False, nullable=True),
                    sa.Column('order_interval_days', sa.NUMERIC(precision=5, scale=1), autoincrement=False, nullable=True),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'vendor_version_pkey')
    )

    # report_code_version
    op.create_table('report_code_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'report_code_version_pkey')
    )

    # family_version
    op.create_table('family_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'family_version_pkey')
    )

    # deposit_link_version
    op.create_table('deposit_link_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('description', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
                    sa.Column('amount', sa.NUMERIC(precision=5, scale=2), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'deposit_link_version_pkey')
    )

    # department_version
    op.create_table('department_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('number', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.VARCHAR(length=30), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'department_version_pkey')
    )

    # subdepartment_version
    op.create_table('subdepartment_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('number', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.VARCHAR(length=30), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'subdepartment_version_pkey')
    )

    # tax_version
    op.create_table('tax_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('description', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
                    sa.Column('rate', sa.NUMERIC(precision=7, scale=5), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'tax_version_pkey')
    )

    # brand_version
    op.create_table('brand_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('name', sa.VARCHAR(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'brand_version_pkey')
    )

    # person_version
    op.create_table('person_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('first_name', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('last_name', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('display_name', sa.VARCHAR(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'person_version_pkey')
    )

    # product_price_version
    op.create_table('product_price_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('product_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('type', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('level', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('starts', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
                    sa.Column('ends', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
                    sa.Column('price', sa.NUMERIC(precision=8, scale=3), autoincrement=False, nullable=True),
                    sa.Column('multiple', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('pack_price', sa.NUMERIC(precision=8, scale=3), autoincrement=False, nullable=True),
                    sa.Column('pack_multiple', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'product_price_version_pkey')
    )

    # user_version
    op.create_table('user_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('username', sa.VARCHAR(length=25), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('active', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'user_version_pkey')
    )

    # customer_version
    op.create_table('customer_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
                    sa.Column('name', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
                    sa.Column('email_preference', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'customer_version_pkey')
    )

    # email_version
    op.create_table('email_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('type', sa.VARCHAR(length=15), autoincrement=False, nullable=True),
                    sa.Column('address', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'email_version_pkey')
    )

    # employee_version
    op.create_table('employee_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('status', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('display_name', sa.VARCHAR(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'employee_version_pkey')
    )

    # address_version
    op.create_table('address_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('type', sa.VARCHAR(length=15), autoincrement=False, nullable=True),
                    sa.Column('street', sa.VARCHAR(length=100), autoincrement=False, nullable=True),
                    sa.Column('street2', sa.VARCHAR(length=100), autoincrement=False, nullable=True),
                    sa.Column('city', sa.VARCHAR(length=60), autoincrement=False, nullable=True),
                    sa.Column('state', sa.VARCHAR(length=2), autoincrement=False, nullable=True),
                    sa.Column('zipcode', sa.VARCHAR(length=10), autoincrement=False, nullable=True),
                    sa.Column('invalid', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'address_version_pkey')
    )

    # store_version
    op.create_table('store_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.VARCHAR(length=10), autoincrement=False, nullable=True),
                    sa.Column('name', sa.VARCHAR(length=100), autoincrement=False, nullable=True),
                    sa.Column('database_key', sa.VARCHAR(length=30), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'store_version_pkey')
    )

    # product_code_version
    op.create_table('product_code_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('product_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('ordinal', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'product_code_version_pkey')
    )

    # customer_group_version
    op.create_table('customer_group_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
                    sa.Column('name', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'customer_group_version_pkey')
    )

    # label_profile_version
    op.create_table('label_profile_version',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('ordinal', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.VARCHAR(length=3), autoincrement=False, nullable=True),
                    sa.Column('description', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('printer_spec', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
                    sa.Column('formatter_spec', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
                    sa.Column('format', sa.TEXT(), autoincrement=False, nullable=True),
                    sa.Column('visible', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BIGINT(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BIGINT(), autoincrement=False, nullable=True),
                    sa.Column('operation_type', sa.SMALLINT(), autoincrement=False, nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id', name=u'label_profile_version_pkey')
    )



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync for Linux
"""

from __future__ import unicode_literals, absolute_import

import sys
import time
import logging
from traceback import format_exception

from rattail.daemon import Daemon
from rattail.threads import Thread
from rattail.datasync.config import load_profiles
from rattail.datasync.util import get_lastrun, get_lastrun_setting, get_lastrun_timefmt, next_batch_id


log = logging.getLogger(__name__)


class DataSyncDaemon(Daemon):
    """
    Linux daemon implementation of DataSync.

    This is normally started via command line, e.g.:

    .. code-block:: sh

       cd /srv/envs/poser
       bin/rattail -c app/datasync.conf datasync start

    .. note::
       Even though the naming implies a "proper" daemon, it will not
       actually daemonize itself.  For true daemon behavior, you
       should run this using a wrapper such as `supervisor`_.

       .. _supervisor: http://supervisord.org
    """

    def run(self):
        """
        Starts watcher and consumer threads according to configuration.

        A separate thread is started for each watcher defined in
        config.  :func:`watch_for_changes()` is the target callable
        for each of these threads.

        Additionally, a separate thread is started for each consumer
        defined for the watcher.  :func:`consume_changes_perpetual()`
        is the target callable for each of these threads.

        Once all threads are started, this (main) thread loops forever
        (so as to stay alive, and hence keep child threads alive) but
        takes no further actions.
        """
        for key, profile in load_profiles(self.config).items():

            # Create watcher thread for the profile.
            name = '{}-watcher'.format(key)
            log.debug("starting thread '{}' with watcher: {}".format(name, profile.watcher_spec))
            thread = Thread(target=watch_for_changes, name=name, args=(self.config, profile.watcher))
            thread.daemon = True
            thread.start()

            # Create consumer threads, unless watcher consumes itself.
            if not profile.watcher.consumes_self:

                # Create a thread for each "isolated" consumer.
                # for consumer in profile.isolated_consumers:
                for consumer in profile.consumers:
                    name = '{}-consumer-{}'.format(key, consumer.key)
                    log.debug("starting thread '%s' with consumer: %s", name, consumer.spec)
                    thread = Thread(target=consume_changes_perpetual, name=name, args=(self.config, consumer))
                    thread.daemon = True
                    thread.start()

        # Loop indefinitely.  Since this is the main thread, the app will
        # terminate when this method ends; all other threads are "subservient"
        # to this one.
        while True:
            time.sleep(.01)


def watch_for_changes(config, watcher):
    """
    Target for datasync watcher threads.

    This function will loop forever (barring an error) and
    periodically invoke the ``watcher`` object to check for any new
    changes, then pause before doing it again, etc.  It also tries to
    detect errors and handle them gracefully.

    The watcher itself is of course responsible for the mechanics of
    checking for new changes.  It also determines how frequently the
    checks should happen, etc.

    Each time the watcher finds/returns changes, this function will
    invoke :func:`record_or_process_changes()` with the result.

    :param config: Config object for the app.

    :param watcher: Reference to datasync watcher object.
    """
    app = config.get_app()

    # let watcher do any setup it needs
    watcher.setup()

    # the 'last run' value is maintained as zone-aware UTC
    lastrun = get_lastrun(config, watcher.key)
    lastrun_setting = get_lastrun_setting(config, watcher.key)
    timefmt = get_lastrun_timefmt(config)

    # outer loop - this should never terminate unless an unhandled
    # exception happens, or the daemon is stopped
    while True:

        # reset for inner loop
        thisrun = app.make_utc(tzinfo=True)
        attempts = 0
        last_error_type = None

        # inner loop - this is for sake of retry
        while True:
            attempts += 1

            try:
                changes = watcher.get_changes(lastrun)

            except Exception as error:

                exc_type, exc, traceback = sys.exc_info()

                # If we've reached our final attempt, stop retrying.
                if attempts >= watcher.retry_attempts:
                    log.warning("attempt #%s of %s failed calling `watcher.get_changes()`, "
                                "this thread will now *terminate* until datasync restart",
                                attempts, watcher.retry_attempts, exc_info=True)
                    app.send_email('datasync_error_watcher_get_changes', {
                        'watcher': watcher,
                        'error': exc,
                        'attempts': attempts,
                        'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
                        'datasync_url': config.datasync_url(),
                    })
                    raise

                # If this exception is not the first, and is of a different type
                # than seen previously, do *not* continue to retry.
                if last_error_type is not None and not isinstance(error, last_error_type):
                    log.exception("new exception differs from previous one(s), "
                                  "giving up on watcher.get_changes()")
                    raise

                # remember which type of error this was; pause for next retry
                last_error_type = type(error)
                log.warning("attempt #%s of %s for watcher.get_changes() failed",
                            attempts, watcher.retry_attempts, exc_info=True)
                log.debug("pausing for %s seconds before making next attempt",
                          watcher.retry_delay)
                if watcher.retry_delay:
                    time.sleep(watcher.retry_delay)

            else: # watcher got changes okay (possibly empty set)

                # record new lastrun time
                lastrun = thisrun
                try:
                    session = app.make_session()
                except:
                    log.exception("failed to make session, to save lastrun time")
                    raise
                else:
                    try:
                        app.save_setting(session, lastrun_setting,
                                         lastrun.strftime(timefmt))
                        session.commit()
                    except:
                        session.rollback()
                        log.exception("failed to save lastrun time")
                        raise
                    finally:
                        session.close()

                if changes:
                    log.debug("got %s changes from watcher", len(changes))

                    # record or process changes (depends on watcher)
                    try:
                        record_or_process_changes(config, watcher, changes)
                    except:
                        log.exception("failed to record/process changes")
                        raise

                    # prune changes if applicable (depends on watcher)
                    try:
                        prune_changes(config, watcher, changes)
                    except:
                        log.exception("failed to prune changes")
                        raise

                # break out of inner loop to reset the attempt count
                # for next grab
                break

        # pause between successful change grabs
        time.sleep(watcher.delay)


def record_or_process_changes(config, watcher, changes):
    """
    This function is responsible for the "initial" handling of changes
    obtained from a watcher.  What this means will depend on the
    watcher, as follows:

    Most watchers are just that - their only job is to report new
    changes to datasync.  In this case this function will merely
    "record" the changes, by inserting them into the
    ``datasync_change`` queue table for further processing by the
    consumer(s).

    But some watchers "consume self" - meaning any changes they find,
    they also are responsible for processing.  In this case this
    function will "process" (consume) the changes directly, by
    invoking the watcher to do so.  These changes will *not* be added
    to the queue table for any other consumer(s) to process.

    :param config: Config object for the app.

    :param watcher: Reference to datasync watcher, from whence changes
       came.

    :param changes: List of changes obtained from the watcher.

    :returns: ``True`` if all goes well, ``False`` if error.

       TODO: Actually current logic will raise an error instead of
       returning ``False``.  That may be preferable, in which case
       docs should be updated.  But technically it does still return
       ``True`` if no error.
    """
    app = config.get_app()
    model = config.get_model()

    # if watcher consumes itself, then it will process its own
    # changes.  note that there are no assumptions made about the
    # format or structure of these change objects.
    if watcher.consumes_self:
        try:
            session = app.make_session()
        except:
            log.exception("failed to make session")
            raise
        try:
            watcher.process_changes(session, changes)
        except:
            log.exception("watcher failed to process its changes")
            session.rollback()
            raise
        else:
            session.commit()
            log.debug("watcher has consumed its own changes")
            return True
        finally:
            session.close()

    # will record changes to consumer queue...

    # give all change stubs the same timestamp, to help identify them
    # as a "batch" of sorts, so consumers can process them as such.
    # (note, this is less important for identifiying a batch now that
    # we have batch_id, but is probably still helpful anyway)
    now = app.make_utc()

    # save change stub records to rattail database, for consumer
    # thread to find and process
    saved = 0
    try:
        session = app.make_session()
    except:
        log.exception("failed to make session for recording changes")
        raise
    try:
        # assign new/unique batch_id so that consumers can keep things
        # straight
        batch_id = next_batch_id(session)
        batch_seq = 0
        for key, change in changes:
            batch_seq += 1
            for consumer in watcher.consumer_stub_keys:
                session.add(model.DataSyncChange(
                    source=watcher.key,
                    batch_id=batch_id,
                    batch_sequence=batch_seq,
                    payload_type=change.payload_type,
                    payload_key=change.payload_key,
                    deletion=change.deletion,
                    obtained=now,
                    consumer=consumer))
                saved += 1
            session.flush()
    except:
        log.exception("failed to record changes")
        session.rollback()
        raise
    else:
        session.commit()
    finally:
        session.close()

    log.debug("saved %s '%s' changes to datasync queue", saved, watcher.key)
    return True


def prune_changes(config, watcher, changes):
    """
    Tell the watcher to prune the original change records from its source
    database, if relevant.
    """
    if not watcher.prunes_changes:
        return

    try:
        # note that we only give it the keys for this
        pruned = watcher.prune_changes([c[0] for c in changes])
    except:
        log.exception("failed to prune changes")
        raise
    if pruned is not None:
        log.debug("watcher pruned %s changes", pruned)


def consume_changes_perpetual(config, consumer):
    """
    Target for datasync consumer threads.

    This function will loop forever (barring an error) and
    periodically invoke the ``consumer`` object to process any changes
    in the queue, then pause before doing it again, etc.  It also
    tries to detect errors and handle them gracefully.

    This function is mostly just the loop itself; it calls
    :func:`consume_current_changes()` during each iteration.

    :param config: Config object for the app.

    :param consumer: Reference to datasync consumer object.
    """
    # tell consumer to do initial setup
    consumer.setup()

    # begin thread perma-loop
    while True:

        # try to consume all current changes
        try:
            result = consume_current_changes(config, consumer)
        except:
            log.exception("failed to consume current changes")
            raise

        if not result:
            # consumption failed, so exit the perma-loop (this thread
            # is now dead)
            break

        # wait 1 sec by default, then look for more changes
        time.sleep(consumer.delay)


def consume_current_changes(config, consumer):
    """
    Consume all changes currently available for the given consumer.

    The datasync queue table will be checked, and if it contains any
    changes applicable to the given consumer, then the consumer will
    be invoked to process the changes.

    If there are no applicable changes in the queue, this function
    will return without taking any real action.  But if there are
    changes, then it tries to be smart about processing them in the
    correct order, as follows:

    The changes are sorted by
    :attr:`~rattail.db.model.datasync.DataSyncChange.obtained` in
    order to determine the earliest timestamp.  Then it calls
    :func:`consume_changes_from()` with that timestamp.

    Once all changes with that timestamp have been processed
    (consumed), this function again looks for any applicable changes
    in the queue, sorting by timestamp and then calling
    :func:`consume_changes_from()` with earliest timestamp.

    This process repeats until there are no longer any changes in the
    queue which pertain to the given consumer.

    :param config: Config object for the app.

    :param consumer: Reference to datasync consumer object.

    :returns: ``True`` if all goes well, ``False`` if error.
    """
    app = config.get_app()
    model = config.get_model()

    try:
        session = app.make_session()
    except:
        log.exception("failed to make session for consuming changes")
        raise

    def get_first_change():
        change = session.query(model.DataSyncChange)\
                        .filter(model.DataSyncChange.source == consumer.watcher.key)\
                        .filter(model.DataSyncChange.consumer == consumer.key)\
                        .order_by(model.DataSyncChange.obtained)\
                        .first()
        return change

    # determine first 'obtained' timestamp
    try:
        first = get_first_change()
    except:
        log.exception("failed to get first change")
        session.close()
        return False

    error = False
    while first:

        # try to consume these changes
        try:
            if not consume_changes_from(config, session, consumer,
                                        first.obtained):
                error = True
        except:
            error = True
            log.exception("failed to consume changes obtained at: %s",
                          first.obtained)

        if error:
            break

        # fetch next 'obtained' timestamp
        try:
            first = get_first_change()
        except:
            log.exception("failed to get next-first change")
            break

    # no more changes! (or perhaps an error)
    session.close()
    return not error


def consume_changes_from(config, session, consumer, obtained):
    """
    Consume all changes which were "obtained" at the given timestamp.

    This fetches all changes from the datasync queue table, which
    correspond to the given consumer and which have an
    :attr:`~rattail.db.model.datasync.DataSyncChange.obtained` value
    matching the one specified.

    There are two possibilities here: either the matching changes are
    part of a "true" batch (i.e. they have a
    :attr:`~rattail.db.model.datasync.DataSyncChange.batch_id` value),
    or not.

    This function therefore first looks for changes which *do* have a
    ``batch_id``.  If found, it then sorts those changes by
    :attr:`~rattail.db.model.datasync.DataSyncChange.batch_sequence`
    to be sure they are processed in the correct order.

    If none of the changes have a ``batch_id`` then this function does
    not sort the changes in any way; they will be processed in
    (presumably) random order.

    In any case, regardless of ``batch_id``, at this point the
    function has identified a set of changes to be processed as a
    "batch" by the consumer.  But the larger the batch, the longer it
    will take for the consumer to process it.  This brings a couple of
    issues:

    If the consumer is a Rattail DB, and data versioning is enabled,
    this may cause rather massive resource usage if too many data
    writes happen.

    Additionally, there is no true "progress indicator" for datasync
    at this time.  A semi-practical way to determine its progress is
    simply to view the queue table and see what if anything it
    contains (when empty, processing is complete).  The set of changes
    being processed here, will be removed from the queue only after
    being processed.  Hence, the larger the batch, the "less granular"
    the "progress indicator" will be.

    To address these issues then, this function may "prune" the set of
    changes such that only so many are processed at a time.

    And finally, we have a (possibly smaller) set of changes to be
    processed.  This function will then ask the consumer to begin a
    new transaction, then process the changes, and ultimately commit
    the transaction.

    Once processing is complete (i.e. assuming no error) then those
    changes are removed from the queue.

    :param config: Config object for the app.

    :param session: Current session for Rattail DB.

    :param consumer: Reference to datasync consumer object.

    :param obtained: UTC "obtained" timestamp for the first change in
       the queue.  This is used to filter existing changes, i.e. we
       only want to process changes with this same timestamp, as they
       are treated as a single "batch".

    :returns: ``True`` if all goes well, ``False`` if error.
    """
    app = config.get_app()
    model = config.get_model()

    # we only want changes "obtained" at the given time.  however, at least
    # until all code has been refactored, we must take two possibilities into
    # account here: some changes may have been given a batch ID, but some may
    # not.  we will prefer those with batch ID, or fall back to those without.
    changes = session.query(model.DataSyncChange)\
                     .filter(model.DataSyncChange.source == consumer.watcher.key)\
                     .filter(model.DataSyncChange.consumer == consumer.key)\
                     .filter(model.DataSyncChange.obtained == obtained)\
                     .filter(model.DataSyncChange.batch_id != None)\
                     .order_by(model.DataSyncChange.batch_id,
                               model.DataSyncChange.batch_sequence)\
                     .all()
    if changes:
        # okay, we got some with a batch ID, now we must prune that list down
        # so that we're only dealing with a single batch
        batch_id = changes[0].batch_id
        changes = [c for c in changes if c.batch_id == batch_id]
    else:
        # no changes with batch ID, so let's get all without ID instead
        changes = session.query(model.DataSyncChange)\
                         .filter(model.DataSyncChange.source == consumer.watcher.key)\
                         .filter(model.DataSyncChange.consumer == consumer.key)\
                         .filter(model.DataSyncChange.obtained == obtained)\
                         .filter(model.DataSyncChange.batch_id == None)\
                         .all()

    # maybe limit size of batch to process.  this can be useful e.g. when large
    # amounts of changes land in the queue with same timestamp, and versioning
    # is also enabled.
    batch_size = config.getint('rattail.datasync', 'batch_size_limit',
                               session=session)
    if batch_size and len(changes) > batch_size:
        changes = changes[:batch_size]

    log.debug("will process %s changes from %s", len(changes),
              app.localtime(obtained, from_utc=True))

    # first retry loop is to begin the transaction
    attempts = 0
    errtype = None
    while True:
        attempts += 1

        try:
            consumer.begin_transaction()

        except Exception as errobj: # processing failed!
            exc_type, exc, traceback = sys.exc_info()

            # if we've reached our final attempt, stop retrying
            if attempts >= consumer.retry_attempts:
                log.warning("attempt #%s failed calling `consumer.begin_transaction()`; "
                            "this thread will now *terminate* until datasync restart",
                            attempts, exc_info=True)
                app.send_email('datasync_error_consumer_process_changes', {
                    'watcher': consumer.watcher,
                    'consumer': consumer,
                    'error': exc,
                    'attempts': attempts,
                    'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
                    'datasync_url': config.datasync_url(session=session),
                })
                return False

            # if this exception is not the first, and is of a different type
            # than seen previously, do *not* continue to retry
            if errtype is not None and not isinstance(errobj, errtype):
                log.exception("new exception differs from previous one(s), "
                              "giving up on consumer.begin_transaction()")
                return False

            # record the type of exception seen; maybe pause before next retry
            errtype = type(errobj)
            log.warning("attempt #%s failed for '%s' -> '%s' consumer.begin_transaction()",
                        attempts, consumer.watcher.key, consumer.key)
            log.debug("pausing for %s seconds before making attempt #%s of %s",
                      consumer.retry_delay, attempts + 1, consumer.retry_attempts)
            if consumer.retry_delay:
                time.sleep(consumer.retry_delay)

        else: # transaction began okay

            # can stop the attempt/retry loop now
            break

    # second retry loop is to process the changes
    attempts = 0
    errtype = None
    while True:

        attempts += 1

        try:
            consumer.process_changes(session, changes)

        except Exception as errobj: # processing failed!
            exc_type, exc, traceback = sys.exc_info()

            try:
                consumer.rollback_transaction()
            except:
                log.exception("consumer failed to rollback transaction")
                return False

            # if we've reached our final attempt, stop retrying
            if attempts >= consumer.retry_attempts:
                log.warning("attempt #%s failed calling `consumer.process_changes()`; "
                            "this thread will now *terminate* until datasync restart",
                            attempts, exc_info=True)
                app.send_email('datasync_error_consumer_process_changes', {
                    'watcher': consumer.watcher,
                    'consumer': consumer,
                    'error': exc,
                    'attempts': attempts,
                    'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
                    'datasync_url': config.datasync_url(session=session),
                })
                return False

            else: # more attempts to be made, but log error for debug
                log.debug("attempt #%s failed calling `consumer.process_changes()`",
                          attempts, exc_info=True)

            # if this exception is not the first, and is of a different type
            # than seen previously, do *not* continue to retry
            if errtype is not None and not isinstance(errobj, errtype):
                log.exception("new exception differs from previous one(s), "
                              "giving up on consumer.process_changes()")
                return False

            # record the type of exception seen; maybe pause before next retry
            errtype = type(errobj)
            log.warning("attempt #%s failed for '%s' -> '%s' consumer.process_changes()",
                        attempts, consumer.watcher.key, consumer.key)
            log.debug("pausing for %s seconds before making attempt #%s of %s",
                      consumer.retry_delay, attempts + 1, consumer.retry_attempts)
            if consumer.retry_delay:
                time.sleep(consumer.retry_delay)

        else: # consumer processed changes okay

            # commit consumer transaction
            try:
                consumer.commit_transaction()
            except:
                log.exception("consumer failed to commit transaction")
                return False

            # delete these changes from datasync queue
            try:
                for i, change in enumerate(changes):
                    session.delete(change)
                    if i % 200 == 0:
                        session.flush()
                session.commit()
            except:
                log.exception("failed to delete changes from queue")
                return False

            # can stop the attempt/retry loop now
            log.debug("processed %s changes", len(changes))
            break

    return True



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
API for Store Models
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy.orm.exc import NoResultFound

from rattail.db import model


# TODO: need to refactor this into the AppHandler, or..?
def get_store(session, key):
    """
    Locate and return a store for the given key, if possible.

    First the key is assumed to be a ``Store.id`` value.  If no matches are
    found, then it looks for a special setting in the database.  If one is
    found, ``get_store()`` is called again with its value.

    :param session: Active database session.

    :param key: Value to use when searching for the store.

    :returns: The :class:`rattail.db.model.Store` instance if found; otherwise
      ``None``.
    """
    # Store.uuid match?
    store = session.query(model.Store).get(key)
    if store:
        return store

    # Store.id match?
    try:
        return session.query(model.Store).filter_by(id=key).one()
    except NoResultFound:
        pass

    # Try settings, if value then recurse.
    from rattail.db.api import get_setting
    key = get_setting(session, 'rattail.store.{0}'.format(key))
    if key:
        return get_store(session, key)
